import numpy as np

from .environments import (
    BaseEnvironment,
    FunctionOptim,
    Genotype,
    ImageOptim,
    MatchString,
    Sodarace,
)

# ----- Generate sample seeds and init args for environments -----
# They are simple template arguments to initialize several environments.
# Sample usage:
#   from openelm.environment import sodarace_init_args
#   sodarace = Sodarace(**sodarace_init_args, run_name="test")


IMAGE_SEED = {
    "program_str": """import numpy as np
def draw_blue_rectangle() -> np.ndarray:
\tpic = np.zeros((32, 32, 3))
\tfor x in range(2, 30):
\t\tfor y in range(2, 30):
\t\t\tpic[x, y] = np.array([0, 0, 255])
\treturn pic
""",
    "result_obj": None,
    "error_code": 0,
}
exec(IMAGE_SEED["program_str"], globals())
IMAGE_SEED["result_obj"] = globals()["draw_blue_rectangle"]()
target = np.zeros((32, 32, 3))
for y in range(32):
    for x in range(32):
        if (y - 16) ** 2 + (x - 16) ** 2 <= 100:  # a radius-10 circle
            target[y, x] = np.array([1, 1, 0])


SQUARE_SEED = {
    "program_str": """from openelm.environments.sodaracer.walker.walk_creator import walker_creator


def make_square(wc, x0, y0, x1, y1):
    \"\"\"Make a square with top left x0,y0 and top right x1,y1.\"\"\"
    j0 = wc.add_joint(x0, y0)
    j1 = wc.add_joint(x0, y1)
    j2 = wc.add_joint(x1, y1)
    j3 = wc.add_joint(x1, y0)
    return j0, j1, j2, j3


def make_walker():
    wc = walker_creator()

    # the main body is a square
    sides = make_square(wc, 0, 0, 10, 10)
    center = wc.add_joint(5, 5)

    # connect the square with distance muscles
    for k in range(len(sides) - 1):
        wc.add_muscle(sides[k], sides[k + 1])
    wc.add_muscle(sides[3], sides[0])

    # one prong of the square is a distance muscle
    wc.add_muscle(sides[3], center)

    # the other prongs from the center of the square are active
    wc.add_muscle(sides[0], center, 5.0, 0.0)
    wc.add_muscle(sides[1], center, 10.0, 0.0)
    wc.add_muscle(sides[2], center, 2.0, 0.0)

    return wc.get_walker()

""",
    "result_obj": {
        "joints": [(0, 0), (0, 10), (10, 10), (10, 0), (5, 5)],
        "muscles": [
            [0, 1, {"type": "distance", "amplitude": 0.0, "phase": 0.0}],
            [1, 2, {"type": "distance", "amplitude": 0.0, "phase": 0.0}],
            [2, 3, {"type": "distance", "amplitude": 0.0, "phase": 0.0}],
            [3, 0, {"type": "distance", "amplitude": 0.0, "phase": 0.0}],
            [3, 4, {"type": "distance", "amplitude": 0.0, "phase": 0.0}],
            [0, 4, {"type": "muscle", "amplitude": 5.0, "phase": 0.0}],
            [1, 4, {"type": "muscle", "amplitude": 10.0, "phase": 0.0}],
            [2, 4, {"type": "muscle", "amplitude": 2.0, "phase": 0.0}],
        ],
    },
    "error_code": 0,
}

# A sample init args for ImageOptim
image_init_args = {
    "seed": IMAGE_SEED,
    "config": "openelm/config/elm_image_cfg.yaml",
    "target_img": target,
    "diff_model": None,
    "behavior_mode": "3-channel",
}

# A sample init args for Sodarace
sodarace_init_args = {
    "seed": SQUARE_SEED,
    "config": "openelm/config/elm_sodarace_cfg.yaml",
    "diff_model": None,
    "eval_steps": 1000,
}

# ----- (Sample init args end) -----

__all__ = [
    "Genotype",
    "BaseEnvironment",
    "FunctionOptim",
    "ImageOptim",
    "MatchString",
    "Sodarace",
    "IMAGE_SEED",
    "image_init_args",
    "SQUARE_SEED",
    "sodarace_init_args",
]



"""
Simulator.py, mostly just https://github.com/OptimusLime/iesor-physics
translated to python

"""

import json
import math
import os

from Box2D import Box2D as b2

from openelm.environments.sodaracer.helpers import (
    Bone,
    DistanceAccessor,
    Entity,
    Muscle,
    PhysicsId,
    staticBodyCount,
)


class IESoRWorld:
    """Class for the Sodarace simulation."""

    def __init__(self):
        self.world = None
        self.interpolation = 0.0
        self.radians = 0.0
        self.accumulator = 0.0
        self.desiredFPS = 0.0
        self.simulation_rate = 0.0
        self.static_body_count: int = 0
        # keep a list of body identifiers
        self.body_list = list()
        # keep a list of shape identifiers
        self.shape_list = list()
        # keep a reference to every body according to ID
        self.body_map = dict()
        # keep track of all the joints inside the system
        self.bone_list = list()
        # we need to keep track of all our muscles too
        self.muscle_list = list()

        #
        #
        # IESoRWorld::IESoRWorld()
        # {
        #     this->interpolation =0;
        self.interpolation = 0
        #     accumulator = 0;
        self.accumulator = 0
        #     desiredFPS = 60;
        self.desiredFPS = 60
        #     simulationRate = 1.0 / desiredFPS;
        self.simulation_rate = 1.0 / self.desiredFPS
        #     radians = 0;
        self.radians = 0
        #
        #
        #     //this->bodyList = new vector<PhysicsID*>();
        #     //this->shapeList = new vector<PhysicsID*>();
        #     //this->bodyList = new Json::Value(Json::arrayValue);
        #     //this->shapeList = new Json::Value(Json::arrayValue);
        #
        #     // Define the gravity vector.
        #     b2Vec2 gravity(0.0f, -15.0f);
        self.gravity = b2.b2Vec2(0.0, -15.0)  # ??? Magic numbers.
        #
        #     // Construct a world object, which will hold and simulate the rigid bodies.
        #     this->world = new b2World(gravity);
        #     this->world->SetAutoClearForces(false);
        self.world = b2.b2World(self.gravity)
        self.world.autoClearForces = False
        #
        #     // Define the ground body.
        #     b2BodyDef groundBodyDef;
        #     groundBodyDef.type = b2_staticBody;
        #     groundBodyDef.position.Set(0.0f, -18.0f);
        self.groundBodyDef = b2.b2BodyDef()
        self.groundBodyDef.type = b2.b2_staticBody
        self.groundBodyDef.position = b2.b2Vec2(0.0, -18.0)
        #
        #     // Call the body factory which allocates memory for the ground body
        #     // from a pool and creates the ground box shape (also from a pool).
        #     // The body is also added to the world.
        #     b2Body* groundBody = this->addBodyToWorld(
        #         "ground", &groundBodyDef);//this->world->CreateBody(&groundBodyDef);
        self.groundBody = self.add_body_to_world("ground", self.groundBodyDef)
        #     //b2Body* groundBody = this->world->CreateBody(&groundBodyDef);
        #
        #
        #     // Define the ground box shape.
        #     b2PolygonShape groundBox;
        self.groundBox = b2.b2PolygonShape()
        #
        #     // The extents are the half-widths of the box.
        #     groundBox.SetAsBox(350.0, 10.0);
        self.groundBox.SetAsBox(350.0, 10.0)
        #
        #     // Add the ground fixture to the ground body.
        #     this->addShapeToBody(groundBody, &groundBox, 0.0f);
        self.add_shape_to_body(self.groundBody, self.groundBox, 0.0)
        #     //groundBody->CreateFixture(&groundBox, 0.0f);
        #
        #     //Pulling in some sample data for testing!
        #     std::string bodyJSON = loadDataFile("sampleBody224632.json");
        # self.body_json = self.load_data_file("sampleBody224632.json")
        #
        #     //pull in our JSON body plz
        #     Json::Value inBody;
        #     Json::Reader read;
        #     read.parse(bodyJSON, inBody, true);
        # self.in_body = json.loads(self.body_json)
        #
        #     b2Vec2 canvas(200, 150);
        self.canvas = b2.b2Vec2(200, 150)
        #     this->loadBodyIntoWorld(inBody, canvas);
        # self.load_body_into_world(self.in_body, self.canvas)
        #
        #
        #     // Define the dynamic body. We set its position and call the body factory.
        #     b2BodyDef bodyDef;
        # self.body_def = b2.b2BodyDef()
        #     bodyDef.type = b2_dynamicBody;
        # self.body_def.type = b2.b2_dynamicBody
        #     bodyDef.position.Set(0.0f, 24.0f);
        # self.body_def.position = b2.b2Vec2(0.0, 24.0)
        #
        #     //add body to world using definition
        #     b2Body* body = this->addBodyToWorld("rect1", &bodyDef);
        # self.body = self.add_body_to_world("rect1", self.body_def)
        #
        #     // Define another box shape for our dynamic body.
        #     b2PolygonShape dynamicBox;
        # self.dynamic_box = b2.b2PolygonShape()
        #     dynamicBox.SetAsBox(5.0f, 5.0f);
        # self.dynamic_box.SetAsBox(5.0, 5.0)
        #
        #     // Define the dynamic body fixture.
        #     b2FixtureDef fixtureDef;
        # self.fixture_def = b2.b2FixtureDef()
        #     fixtureDef.shape = &dynamicBox;
        # self.fixture_def.shape = self.dynamic_box
        #     fixtureDef.restitution = .5;
        # self.fixture_def.restitution = .5
        #
        #     // Set the box density to be non-zero, so it will be dynamic.
        #     fixtureDef.density = 1.0f;
        # self.fixture_def.density = 1.0
        #
        #     // Override the default friction.
        #     fixtureDef.friction = 0.3f;
        # self.fixture_def.friction = 0.3
        #
        #     // Add the shape to the body.
        #     this->addShapeToBody(body, &fixtureDef);
        # self.add_shape_to_body_fixture(self.body, self.fixture_def)
        #
        #     // Define the circle body. We set its position and call the body factory.
        #     b2BodyDef cDef;
        # self.c_def = b2.b2BodyDef()
        #     cDef.type = b2_dynamicBody;
        # self.c_def.type = b2.b2_dynamicBody
        #     cDef.position.Set(10.0f, 24.0f);
        # self.c_def.position = b2.b2Vec2(10.0, 24.0)
        #
        #     //add body to world using definition
        #     body = this->addBodyToWorld("circleTest", &cDef);
        # self.body = self.add_body_to_world("circleTest", self.c_def)
        #
        #     // Define another box shape for our dynamic body.
        #     b2CircleShape dCircle;
        # self.d_circle = b2.b2CircleShape()
        #     dCircle.m_radius = 5.0;
        # self.d_circle.radius = 5.0
        #
        #     // Define the dynamic body fixture.
        #     b2FixtureDef circleDef;
        # self.circle_def = b2.b2FixtureDef()
        #     circleDef.shape = &dCircle;
        # self.circle_def.shape = self.d_circle
        #     circleDef.restitution = .5;
        # self.circle_def.restitution = 0.5
        #
        #     // Set the box density to be non-zero, so it will be dynamic.
        #     circleDef.density = 1.0f;
        # self.circle_def.density = 1.0
        #
        #     // Override the default friction.
        #     circleDef.friction = 0.3f;
        # self.circle_def.friction = 0.3
        #
        #     // Add the shape to the body.
        #     this->addShapeToBody(body, &circleDef);
        # self.add_shape_to_body_fixture(self.body, self.circle_def)
        #
        #
        #
        #     //Add some forces!
        #     //body->ApplyAngularImpulse(.4, true);
        #
        #     body->ApplyTorque(150, true);
        # self.body.ApplyTorque(150, True)
        #     b2Vec2 pulse(70, 0);
        #     b2Vec2 o(0,3);
        #     body->ApplyLinearImpulse(pulse, o, true);
        # self.body.ApplyLinearImpulse(
        #     b2.b2Vec2(70, 0), b2.b2Vec2(0, 3), True
        # )
        #
        # }

    def worldDrawList(self) -> str:
        # string IESoRWorld::worldDrawList()
        # {
        #     //This will be written to json
        #     //fastwriter is not human readable --> compact
        #     //exactly what we want to sent to our html files
        #     Json::FastWriter* writer = new Json::FastWriter();
        #
        #     //root is what we'll be sending back with all the shapes (stored in shapes!)
        #     Json::Value root;
        root = dict()
        #     Json::Value shapes;
        shapes = dict()
        #
        #     //we'll loop through all required body information
        #     b2Body * B = this->world->GetBodyList();
        b_list = self.world.bodies
        #
        #     while(B != NULL)
        for b in b_list:  # type: b2.b2Body
            userData = b.userData
            if userData is not None:
                #     {
                #         if(B->GetUserData())
                #         {
                #             //fetch our body identifier
                #             PhysicsID* pid = static_cast<PhysicsID*>(B->GetUserData());//*((string*)B->GetUserData());
                pid = userData  # type: PhysicsId
                #             std::string bodyID = pid->ID();
                body_id = pid.id()

                #
                #             //we must get all our shapes
                #             b2Fixture* F = B->GetFixtureList();
                f_list = b.fixtures
                #
                #             //cycle through the shapes
                #             while(F != NULL)
                #             {
                for f in f_list:  # type: b2.b2Fixture
                    #                 //Hold our shape drawing information
                    #                 Json::Value singleShape;
                    single_shape = dict()
                    #
                    #                 switch (F->GetType())
                    #                 {
                    #                 case b2Shape::e_circle:
                    if f.type == b2.b2Shape.e_circle:
                        #                     {
                        #                         b2CircleShape* circle = (b2CircleShape*) F->GetShape();
                        circle = f.shape  # type: b2.b2CircleShape
                        #                         /* Do stuff with a circle shape */
                        #                         Json::Value center = positionToJSONValue(circle->m_p);
                        center = position_to_json_value(circle.pos)
                        #                         Json::Value radius = circle->m_radius;
                        radius = circle.radius
                        #                         singleShape["type"] = "Circle";
                        single_shape["type"] = "circle"
                        #                         singleShape["bodyOffset"] = positionToJSONValue(B->GetPosition());
                        single_shape["bodyOffset"] = position_to_json_value(b.position)
                        #                         singleShape["rotation"] = B->GetAngle();
                        single_shape["rotation"] = b.angle
                        #                         singleShape["center"] = center;
                        single_shape["center"] = center
                        #                         singleShape["radius"] = circle->m_radius;
                        single_shape["radius"] = radius
                        #                         singleShape["color"] = "#369";
                        single_shape["color"] = "#369"
                    #                     }
                    #                     break;
                    elif f.type == b2.b2Shape.e_polygon:
                        #                 case b2Shape::e_polygon:
                        #                     {
                        #                         b2PolygonShape* poly = (b2PolygonShape*) F->GetShape();
                        poly = f.shape  # type: b2.b2PolygonShape
                        #                         /* Do stuff with a polygon shape */
                        #
                        #                         Json::Value points = listOfPoints(poly->m_vertices, poly->m_count);
                        #                         singleShape["type"] = "Polygon";
                        single_shape["type"] = "polygon"
                        #                         singleShape["points"] = points;
                        single_shape["points"] = list_of_points(
                            poly.vertices, poly.vertexCount
                        )
                        #                         singleShape["bodyOffset"] = positionToJSONValue(B->GetPosition());
                        single_shape["bodyOffset"] = position_to_json_value(b.position)
                        #                         singleShape["rotation"] = B->GetAngle();
                        single_shape["rotation"] = b.angle
                        #                         singleShape["color"] = "#38F";
                        single_shape["color"] = "#38F"
                    #
                    #                     }
                    #                     break;
                    #                 }
                    #
                    #                 //Each shape is the unique combination of
                    #                 pid = static_cast<PhysicsID*>(F->GetUserData());//*((string*)B->GetUserData());
                    pid = f.userData
                    #
                    #                 string shapeID = pid->ID();// *((string*)F->GetUserData());
                    shape_id = pid.id()
                    #
                    #                 string fullID = bodyID + "_" + shapeID;
                    full_id = body_id + shape_id
                    #                 shapes[fullID] = singleShape;
                    shapes[full_id] = single_shape
        #
        #                 F = F->GetNext();
        #             }
        #         }
        #
        #         B = B->GetNext();
        #     }
        #     //we set our shapes using the body loops
        #     root["shapes"] = shapes;
        root["shapes"] = shapes
        #
        #     //we now need to process all of our joints as well
        #     b2Joint * J = this->world->GetJointList();
        j_list = self.world.joints
        #
        #     Json::Value joints;
        joints = dict()
        #
        #     while(J != NULL)
        for j in joints:  # type: b2.b2Joint
            #     {
            #         if(J->GetUserData())
            if j.userData is not None:
                #         {
                #             //fetch our body identifier
                #             Bone* jid = static_cast<Bone*>(J->GetUserData());
                jid = j.userData  # type: Bone
                #
                #             //we grab the joint identifier
                #             std::string bodyID = jid->ID();
                body_id = jid.id()
                #
                #             //Hold our joint drawing information
                #             Json::Value singleJoint;
                single_joint = dict()
                #
                #             //we should use the body identifiers
                #             //but they both need to exist for this to be valid
                #             if(J->GetBodyA()->GetUserData() && J->GetBodyB()->GetUserData())
                if (j.bodyA.userData is not None) and (j.bodyB.userData is not None):
                    #             {
                    #                 //we need to know what bodies are connected
                    #                 PhysicsID* pid = static_cast<PhysicsID*>(J->GetBodyA()->GetUserData());
                    pid = j.bodyA.userData  # type: PhysicsId
                    #                 singleJoint["sourceID"] =  pid->ID();
                    single_joint["sourceID"] = pid.id()
                    #
                    #                 pid = static_cast<PhysicsID*>( J->GetBodyB()->GetUserData());
                    pid = j.bodyB.userData  # type: PhysicsId
                    #                 singleJoint["targetID"] =  pid->ID();
                    single_joint["targetID"] = pid
                    #
                    #                 //now we need more drawing informtion regarding offsets on the body
                    #
                    #                 //get the anchor relative to body A
                    #                 //set that in our json object
                    #                 singleJoint["sourceOffset"] = positionToJSONValue(J->GetAnchorA());
                    single_joint["sourceOffset"] = position_to_json_value(j.anchorA)
                    #                 //set the same for our second object
                    #                 singleJoint["targetOffset"] = positionToJSONValue(J->GetAnchorB());
                    single_joint["targetOffset"] = position_to_json_value(j.anchorB)
                #             }
                #
                #             //set our joint object using the id, and json values
                #             joints[bodyID] = singleJoint;
                joints[body_id] = single_joint
        #         }
        #
        #         //loop to the next object
        #         J = J->GetNext();
        #     }
        #
        #     root["joints"] = joints;
        root["joints"] = joints
        #
        #     return writer->write(root);
        # }
        return json.dumps(root)

    def updateWorld(self, ms_update: float) -> int:
        #
        #
        # //pass in the amount of time you want to simulate
        # //in real time, this is the delta time between called to update
        # //in simulation time, this can be any amount -- but this function won't return until all updates are processed
        # //so it holds up the thread
        # int IESoRWorld::updateWorld(double msUpdate)
        # {
        #     msUpdate = 4*msUpdate;
        ms_update = 4 * ms_update  # why...
        #
        #     //we'll keep track of the number of times we simulated during this update
        #     int stepCount = 0;
        step_count: int = 0
        #
        #     //# of seconds since last time
        #     double frameTime = msUpdate/1000.0f;
        frame_time = ms_update / 1000.0
        #
        #     //maximum frame time, to prevent what is called the spiral of death
        #     if (frameTime > .35)
        #         frameTime = .35;
        frame_time = 0.35 if frame_time > 0.35 else frame_time
        #
        #     //we accumulate all the time we haven't rendered things in
        #     this->accumulator += frameTime;
        self.accumulator += frame_time
        #
        #     //as long as we have a chunk of time to simulate, we need to do the simulation
        #     while (accumulator >= simulationRate)
        #     {
        while self.accumulator >= self.simulation_rate:
            #         //how many times did we run this loop, we shoudl keep track --
            #         //that's the number of times we ran the physics engine
            #         stepCount++;
            step_count += 1
            #
            #         //move the muscles quicker using this toggle
            #         float speedup = 3;
            speedup = 3
            #
            #         //we loop through all our muscles, and update the lengths associated with the connectionsj
            #         for (std::vector<Muscle*>::iterator it = muscleList.begin() ; it != muscleList.end(); ++it)
            #         {
            for muscle in self.muscle_list:
                #             //grab our muscle pointer from the list
                #             Muscle* muscle = *it;
                #
                #             //get our distance joint -- holder of physical joint info
                #             b2DistanceJoint* dJoint = (b2DistanceJoint*)muscle->GetJoint();
                d_joint = muscle.get_joint()
                #
                #             //Javascript version
                #             //muscle.SetLength(muscle.m_length + muscle.amplitude/this.scale*Math.cos(rad + muscle.phase*2*Math.PI));
                #             //double lengthCalc = (dJoint->GetLength() + muscle->GetAmplitude()*cos(radians + muscle->GetPhase() * 2 * M_PI));
                #
                #             //fetch the original length of the distance joint, and add some fraction of that amount to the length
                #             //depending on the current location in the muscle cycle
                #             double lengthCalc = (1.0 + muscle->GetAmplitude() * cos(radians + muscle->GetPhase() * 2 * M_PI)) * muscle->GetOriginalLength();
                length_calc = (
                    1
                    + muscle.get_amplitude()
                    * math.cos(self.radians + muscle.get_phase() * 2 * math.pi)
                    * muscle.get_original_length()
                )
                #
                #             //we set our length as the calculate value
                #             dJoint->SetLength(lengthCalc);
                DistanceAccessor.setLength(d_joint, length_calc)
            #         }
            #
            #         //step the physics world
            #         this->world->Step(
            #             this->simulationRate   //frame-rate
            #             , 10       //velocity iterations
            #             , 10       //position iterations
            #         );
            self.world.Step(self.simulation_rate, 10, 10)
            #
            #         //manually clear forces when doing fixed time steps,
            #         //NOTE: that we disabled auto clear after step inside of the b2World
            #         this->world->ClearForces();
            self.world.ClearForces()
            #
            #         //increment the radians for the muscles
            #         radians += speedup * this->simulationRate;
            self.radians += speedup * self.simulation_rate
            #
            #         //decrement the accumulator - we ran a chunk just now!
            #         accumulator -= this->simulationRate;
            self.accumulator -= self.simulation_rate
        #     }
        #
        #     //interpolation is basically a measure of how far into the next step we should have simulated
        #     //it's meant to ease the drawing stages
        #     //(you linearly interpolate the last frame and the current frame using this value)
        #     this->interpolation = accumulator / this->simulationRate;
        self.interpolation = self.accumulator / self.simulation_rate
        #
        #     //how many times did we run the simulator?
        #     return stepCount;
        # }
        return step_count

    def add_body_to_world(self, body_id: str, body_def: b2.b2BodyDef) -> b2.b2Body:
        # b2Body* IESoRWorld::addBodyToWorld(string bodyID, b2BodyDef* bodyDef)
        # {
        #     //add body to world
        #     b2Body* body = this->world->CreateBody(bodyDef);
        body = self.world.CreateBody(body_def)  # type: b2.b2Body
        #
        #     //assign the bodyID to a new struct object
        #     PhysicsID* pid = new PhysicsID(bodyID);
        pid = PhysicsId(body_id)
        #
        #     //track all the struct objects
        #     this->bodyList.push_back(pid);
        self.body_list.append(pid)
        #
        #     //track all body objects
        #     this->bodyMap[bodyID] = body;
        self.body_map[body_id] = body
        #
        #     //identify body with user data
        #     body->SetUserData(pid);
        body.userData = pid
        #
        #     return body;
        # }
        return body

    def add_shape_to_body_fixture(
        self, body: b2.b2Body, fix_def: b2.b2FixtureDef
    ) -> b2.b2Fixture:
        # b2Fixture* IESoRWorld::addShapeToBody(b2Body* body, b2FixtureDef* fixDef)
        # {
        #     //create physics id based on global shape count
        #     PhysicsID* pid = createShapeID();
        #
        #     //add the fixture to our body using the definition
        #     b2Fixture* fix = body->CreateFixture(fixDef);
        #
        #     //Attach this info to the shape for identification
        #     setShapeID(fix, pid);
        #
        #     //send back created fixture
        #     return fix;
        # }
        pid = self.create_shape_id()
        fix = body.CreateFixture(fix_def)  # type: b2.b2Fixture
        set_shape_id(fix, pid)
        return fix

    def add_shape_to_body(
        self, body: b2.b2Body, shape: b2.b2Shape, density: float
    ) -> b2.b2Fixture:
        # b2Fixture* IESoRWorld::addShapeToBody(b2Body*body, b2Shape* shape, float density)
        # {
        #     //create physics id based on global shape count
        #     PhysicsID* pid = createShapeID();
        #
        #     //add the fixture to our body,
        #     b2Fixture* fix = body->CreateFixture(shape, density);
        #
        #     //Attach this info to the shape for identification
        #     setShapeID(fix, pid);
        #
        #     //send back created fixture
        #     return fix;
        # }
        pid = self.create_shape_id()
        fix = body.CreateFixture(shape=shape, density=density)  # type: b2.b2Fixture
        set_shape_id(fix, pid)
        return fix

    def load_body_into_world(self, in_body: dict, width_height: b2.b2Vec2) -> dict:
        #
        # std::map<std::string, double>* IESoRWorld::loadBodyIntoWorld(Json::Value& inBody, b2Vec2 widthHeight)
        # {
        # //we use bodycount as a way to identify each physical object being added from the body information
        # int oBodyCount = world->GetBodyCount();
        o_body_count = self.world.bodyCount
        #
        # //This mapping will be what we return from the function while adding the body
        # //we measure the morphological properties
        # map<string, double> morphology;
        morphology = dict()
        #
        # //here we use our body information to create all the necessary body additions to the world
        # vector<Json::Value> entities = getBodyEntities(inBody, widthHeight, morphology);
        entities = self.get_body_entities(in_body, width_height, morphology)
        #
        # //we'll quickly map and ID into a json point value
        # map<string, Json::Value> entityMap;
        entity_map = dict()
        #
        # //this was we can access locations by the bodyID while determining connection distance
        # for (std::vector<Json::Value>::iterator it = entities.begin() ; it != entities.end(); ++it)
        # {
        #     Json::Value e = *it;
        #     entityMap[e["bodyID"].asString()] = e;
        # }
        for e in entities:
            entity_map[e["bodyID"]] = e
        #
        # //push our bodies into the system so that our joints have bodies to connect to
        # this->setBodies(&entities);
        self.set_bodies(entities)
        #
        # //Did we use LEO to calculate our body information -- affects the indexing procedure
        # bool useLEO = inBody["useLEO"].asBool();
        # useLEO = in_body["useLEO"]
        #
        # //this is the json connection array
        # Json::Value connections = inBody["connections"];
        connections = in_body["muscles"]
        #
        # //this determines if we should be a fixed connection (bone) or a moving connection (muscle)
        # double amplitudeCutoff = .2;
        amplitude_cutoff = 0.2
        #
        # //we like to measure the total connection distance as part of our mass calculation
        # //mass = # of nodes + length of connections
        # double connectionDistanceSum = 0.0;
        connection_distance_sum = 0.0
        #
        # //loop through all our connections
        # for(int i=0; i < connections.size(); i++)
        for i in range(len(connections)):
            # {
            #     //grab connection from our array
            #     Json::Value connectionObject = connections[i];
            #     Json::Value cppnOutputs = connectionObject["cppnOutputs"];
            muscle = connections[i]
            # cppn_outputs = muscle["cppnOutputs"]
            #
            #     int sID = atoi(connectionObject["sourceID"].asString().c_str());
            s_id = muscle[0]
            t_id = muscle[1]
            #     int tID = atoi(connectionObject["targetID"].asString().c_str());
            #
            #     //To identify a given object in the physical world, we need to start with the current body count, and add the source id number
            #     //This allows us to add multiple bodies to the same world (though this is recommended against, since it's easier to create separate worlds)
            #     string sourceID = toString(oBodyCount + sID);
            #     string targetID = toString(oBodyCount + tID);
            source_id = str(o_body_count + s_id)
            target_id = str(o_body_count + t_id)
            #
            #      //we ignore connections that loop to themselves
            #     if (sourceID == targetID)
            #     {
            #         continue;
            #     }
            if source_id == target_id:
                continue
            #
            #
            #     try
            #     {
            try:  # oh boy...
                #         //depending on whether LEO was used or not, dictates what outputs we'll be looking at for what
                #         int phaseIx = (useLEO ? 2 : 1);
                #         int ampIx = (useLEO ? 3 : 2);
                # phase_ix = 1
                # amp_ix = 2
                #
                #         //sample the amplitude output to know what's up -- we convert from [-1,1] to [0,1]
                #         double amp = (cppnOutputs[ampIx].asDouble() + 1) / 2;
                amp = (muscle[-1]["amplitude"] + 1) / 2.0
                #
                #         //what's the distance been the source (x,y) and distance (x,y) -- that's the length of our connection
                #         double connectionDistance = sqrt(
                #             pow(entityMap[sourceID]["x"].asDouble() - entityMap[targetID]["x"].asDouble(), 2)
                #             + pow(entityMap[sourceID]["y"].asDouble() - entityMap[targetID]["y"].asDouble(), 2));
                #
                connection_distance = math.sqrt(
                    (entity_map[source_id]["x"] - entity_map[target_id]["x"]) ** 2
                    + (entity_map[source_id]["y"] - entity_map[target_id]["y"]) ** 2
                )
                #         //add the connection distance to our sum
                #         connectionDistanceSum += connectionDistance;
                connection_distance_sum += connection_distance
                #
                #         //this will hold our custom props for the distance/muscle joints
                #         Json::Value props;
                props: dict = dict()
                #
                #         if (amp < amplitudeCutoff)
                #             this->addDistanceJoint(sourceID, targetID, props);//, {frequencyHz: 3, dampingRatio:.3});
                if amp < amplitude_cutoff:
                    self.add_distance_joint(source_id, target_id, props)
                else:
                    #         else{
                    #
                    #             //these are our hardcoded spring behaviors, could be altered by CPPN, but that seems risky
                    #             props["frequencyHz"] = 3;
                    #             props["dampingRatio"] = .3;
                    props["frequencyHz"] = 3
                    props["dampingRatio"] = 0.3
                    #
                    #             //Phase/Amplitude set by our cppn outputs
                    #             props["phase"] = cppnOutputs[phaseIx].asDouble();
                    props["phase"] = muscle[-1]["phase"]
                    #             //JS Version
                    #             //props["amplitude"] = .3f*connectionDistance*amp;
                    #             props["amplitude"] = .3f*amp;
                    props["amplitude"] = 0.3 * amp
                    #
                    #             //need to scale joints based on size of the screen - this is a bit odd, but should help multiple sizes behave the same!
                    #             this->addMuscleJoint(sourceID, targetID, props);
                    self.add_muscle_joint(source_id, target_id, props)
            #         }
            #     }
            #     catch (exception e)
            #     {
            #         printf("Oops error %s", e.what());
            #         throw e;
            #     }
            except Exception as e:
                # why?
                print("Oops error", e)
                raise e
        #
        # }
        #
        # //add the concept of mass
        # morphology["mass"] = morphology["totalNodes"] + connectionDistanceSum/2.0f;
        morphology["mass"] = morphology["totalNodes"] + connection_distance_sum / 2.0
        #
        # //remove the concept of totalNodes since we have mass
        # //morphology should now have width/height, startX/startY, and mass
        # morphology.erase("totalNodes");
        del morphology["totalNodes"]
        #
        # return &morphology;
        return morphology
        # }

    def get_body_entities(
        self, in_body: dict, width_height: b2.b2Vec2, initial_morphology: dict
    ) -> list:
        #
        #
        # std::vector<Json::Value> IESoRWorld::getBodyEntities(Json::Value& inBody, b2Vec2 widthHeight, std::map<std::string, double>& initMorph)
        # {
        #     //We're going to first convert our body into a collection of physics objects that need to be added to the environment
        #     std::vector<Json::Value> entityVector;
        entity_vector = list()
        #
        #     //We'll be using our window width/height for adjusting relative body location
        #     //remember that the CPPN is going to paint information onto a grid, we need to interpret that information according the appropriate sizes
        #     double canvasWidth = widthHeight.x;
        canvas_width = width_height.x
        #     double canvasHeight = widthHeight.y;
        canvas_height = width_height.y
        #
        #     //Grab an array of nodes from our json object
        #     Json::Value oNodes = inBody["nodes"];
        o_nodes = in_body["joints"]
        #
        #     //this is the starting body identifier for each node (every node has a unique ID for the source connections)
        #     //so node 5 is always node 5, but if you already have objects in the physical world
        #     //object 5 != node 5, so the number of objects + nodeID = true nodeID
        #     int bodyCount = this->world->GetBodyCount();
        body_count = self.world.bodyCount
        #     int bodyID = bodyCount;
        body_id = body_count
        #
        #     bool useLEO = inBody["useLEO"].asBool();
        # useLEO = in_body["useLEO"]
        #
        #     //manipulated values of each node, adjusted for screen coorindates
        #     double xScaled, yScaled;
        x_scaled = 0.0
        y_scaled = 0.0
        #
        #     //Adjusting the total size of the object according to window size
        #     double divideForMax = 2.2;
        divide_for_max = 2.2
        #     double divideForMaxHeight = 2.5f;
        divide_for_max_height = 2.5
        #
        #     //Part of a calculation for determinig the x,y values of each node
        #     double maxAllowedWidth = canvasWidth / divideForMax;
        max_allowed_width = canvas_width / divide_for_max
        #     double maxAllowedHeight = canvasHeight / divideForMaxHeight;
        max_allowed_height = canvas_height / divide_for_max_height
        #
        #     //starting values, these will be adjusted
        #     double minX = canvasWidth; double maxX = 0.0f;
        min_x = canvas_width
        max_x = 0.0
        #     double minY = canvasHeight; double maxY = 0.0f;
        min_y = canvas_height
        max_y = 0.0
        #
        #     //We loop through all of our nodes
        #     for (int i=0; i < oNodes.size(); i++)
        for i in range(len(o_nodes)):
            #     {
            #         //for each our nodes, we have a location in the world
            #         Json::Value nodeLocation = oNodes[i];
            node_location = o_nodes[i]
            #
            #         //We pull
            #         double nodeX = nodeLocation["x"].asDouble();
            #         double nodeY = nodeLocation["y"].asDouble();
            node_x = node_location[0]
            node_y = node_location[1]
            #
            #         //here we actually modify the x and y values to fit into a certain sized box depending on the initial screen size/physics world size
            #         xScaled = (nodeX)/divideForMax* maxAllowedWidth;
            x_scaled = node_x / divide_for_max * max_allowed_width
            #         yScaled = (1.0f - nodeY) / divideForMaxHeight* maxAllowedHeight;
            y_scaled = (1.0 - node_y) / divide_for_max_height * max_allowed_height
            #
            #         //FOR each node, we make a body with certain properties, then increment count
            #         entityVector.push_back(Entity::CircleEntity(toString(bodyID), xScaled, yScaled, maxAllowedWidth/35.0, 0));
            entity_vector.append(
                Entity.circle_entity(
                    str(body_id), x_scaled, y_scaled, max_allowed_width / 35.0, 0
                )
            )
            #
            #         ///what's the min/max xValues we've seen? We can determine how wide the creature is
            #         minX = min(minX, xScaled);
            min_x = min(min_x, x_scaled)
            #         maxX = max(maxX, xScaled);
            max_x = max(max_x, x_scaled)
            #
            #         ///what's the min/max y values we've seen? We can determine how tall the creature is
            #         minY = min(minY, yScaled);
            min_y = min(min_y, y_scaled)
            #         maxY = max(maxY, yScaled);
            max_y = max(max_y, y_scaled)
            #
            #         //need to increment the body id so we don't overwrite previous object
            #         bodyID++;
            body_id += 1
        #     }
        #
        #     //We use movex, movey to center our creature in world space after creation
        #     double moveX = (maxX - minX) / 2.0;
        move_x = (max_x - min_x) / 2.0
        #     double moveY = (maxY - minY) / 2.0;
        move_y = (max_y - min_y) / 2.0
        #
        #     //Need to move x,y coordinates for entities
        #     //no one should get an unfair advantage being at a certain screen location
        #     for (std::vector<Json::Value>::iterator it = entityVector.begin() ; it != entityVector.end(); ++it)
        #     {
        for e in entity_vector:
            #         Json::Value e = *it;
            #         e["x"] = (e["x"].asDouble() - minX) + canvasWidth/2 - moveX;
            #         e["y"] = (e["y"].asDouble() - minY) + canvasHeight/2 - moveY;
            e["x"] = (float(e["x"]) - min_x) + canvas_width / 2 - move_x
            e["y"] = (float(e["y"]) - min_y) + canvas_height / 2 - move_y
        #     }
        #
        #     //We now have everything we need to identify our initial morphology
        #     //we can see the inital width/height of the object
        #     //as well as the offset (startx, starty) of the object
        #     initMorph["width"] = maxX - minX;
        initial_morphology["width"] = max_x - min_x
        #     initMorph["height"] = maxY - minY;
        initial_morphology["height"] = max_y - min_y
        #     initMorph["startX"] = minX;
        initial_morphology["startX"] = min_x
        initial_morphology["offsetX"] = -min_x + canvas_width / 2 - move_x
        #     initMorph["startY"] = minY;
        initial_morphology["startY"] = min_y
        initial_morphology["offsetY"] = -min_y + canvas_height / 2 - move_y
        #     initMorph["totalNodes"] = oNodes.size();
        initial_morphology["totalNodes"] = len(o_nodes)
        #
        #     return entityVector;
        # }
        return entity_vector

    def set_bodies(self, entities: list) -> None:
        #
        # //pretty simple, we loop through and call the set body function that does all the work
        # void IESoRWorld::setBodies(vector<Json::Value>* entities)
        # {
        #     //look through our vector of entities, and add each entity to the world
        #     for (std::vector<Json::Value>::iterator it = entities->begin() ; it != entities->end(); ++it)
        #     {
        #         Json::Value e = *it;
        #         string bodyID = e["bodyID"].asString();
        #
        #         this->setBody(e);
        #     }
        # }
        for e in entities:
            body_id = e["bodyID"]
            self.set_body(e)
        return

    def load_data_file(self, data_name: str) -> str:
        #
        # //laod object from data folder -- simple!!
        # std::string IESoRWorld::loadDataFile(std::string dataName)
        # {
        #     std::string filePath = "../../../Physics/Data/" + dataName;
        #     return get_file_contents(filePath.c_str());
        # }

        # TODO: figure out where this should be located. For now this is fine.
        physics_datapath = "Physics/Data"
        return get_file_contents(os.path.join(physics_datapath, data_name))

    def add_distance_joint(self, source_id: str, target_id: str, props: dict) -> Bone:
        #
        #
        # Bone* IESoRWorld::addDistanceJoint(std::string sourceID, std::string targetID, Json::Value props)
        # {
        # //we're connecting two body objects together
        # //they should be defined by their IDs in our bodymap
        # b2Body* body1 = this->bodyMap[sourceID];
        # b2Body* body2 = this->bodyMap[targetID];
        body1 = self.body_map[source_id]  # type: b2.b2Body
        body2 = self.body_map[target_id]  # type: b2.b2Body
        #
        # //Create a basic distance joint
        # b2DistanceJointDef* joint = new b2DistanceJointDef();
        joint = b2.b2DistanceJointDef()
        #
        # //initialize the joints where they're attached at the center of the objects
        # joint->Initialize(body1, body2, body1->GetWorldCenter(), body2->GetWorldCenter());
        joint.Initialize(body1, body2, body1.worldCenter, body2.worldCenter)
        #
        # //check if we defined frequencyHz value, if so fetch it from json
        # //this has to do with how the distance joint responds to stretching
        # if(!props["frequencyHz"].isNull())
        # {
        #     joint->frequencyHz = props["frequencyHz"].asDouble();
        # }
        if props.get("frequencyHz", None) is not None:
            joint.frequencyHz = props["frequencyHz"]
        #
        # //check if we defined dampingRatio value, if so fetch it from json -- check docs for dampingRatio usage
        # if(!props["dampingRatio"].isNull())
        # {
        #     joint->dampingRatio = props["dampingRatio"].asDouble();
        # }
        if props.get("dampingRatio", None) is not None:
            joint.dampingRatio = props["dampingRatio"]
        #
        # //let's use our definition to make a real live joint!
        # b2Joint* wJoint = world->CreateJoint(joint);
        w_joint = self.world.CreateJoint(joint)
        #
        # //we identify our bones by the count in the list
        # Bone* bone = new Bone(toString(boneList.size()), wJoint);
        bone = Bone(str(len(self.bone_list)), w_joint)
        #
        # //we push our joint into a list of joints created
        # boneList.push_back(bone);
        self.bone_list.append(bone)
        #
        # //we store this distance joint info inside of the joint!
        # wJoint->SetUserData(bone);
        w_joint.userData = bone
        #
        # //finished, return our bone object created
        # return bone;
        # }
        #
        #
        return bone

    def add_muscle_joint(self, source_id: str, target_id: str, props: dict) -> Muscle:
        # Muscle* IESoRWorld::addMuscleJoint(std::string sourceID, std::string targetID, Json::Value props)
        # {
        # //we add a standard distance joint
        # Bone* addedJoint = addDistanceJoint(sourceID, targetID, props);
        added_joint = self.add_distance_joint(source_id, target_id, props)
        #
        # //But our muscles have phase and amplitude, which adjust length during updates
        # double phase = 0;
        phase = 0
        #
        # //check if we defined phase value, if so fetch it from json
        # if(!props["phase"].isNull())
        # {
        #     phase = props["phase"].asDouble();
        # }
        if props.get("phase", None) is not None:
            phase = props["phase"]
        #
        # //amplitude is how much change each muscles exhibits
        # double amplitude = 1;
        #
        # //check if it's in our props object
        # if(!props["amplitude"].isNull())
        # {
        #     phase = props["amplitude"].asDouble();
        # }
        amplitude = 1 if props.get("amplitude", None) is None else props["amplitude"]
        #
        # //muscle is just a container for our
        # Muscle* ms = new Muscle(toString(this->muscleList.size()), addedJoint->GetJoint(), amplitude, phase);
        ms = Muscle(str(len(self.muscle_list)), added_joint.joint, amplitude, phase)
        #
        # //track the muscle objects this way, pretty easy
        # this->muscleList.push_back(ms);
        self.muscle_list.append(ms)
        #
        # //send back the joint we added to physical world
        # return ms;
        return ms

    def set_body(self, entity: dict) -> None:
        #
        #
        # void IESoRWorld::setBody(Json::Value& entity)
        # {
        # //we create a generic body def, to hold our body
        # b2BodyDef* bodyDef = new b2BodyDef();
        body_def = b2.b2BodyDef()
        #
        # //we are either adding "ground" (static) or we're adding everything else (dynamic)
        # if (entity["bodyID"].asString() == "ground")
        if entity["bodyID"] == "ground":
            # {
            #     bodyDef->type = b2BodyType::b2_staticBody;
            body_def.type = b2.b2_staticBody
        # }
        # else
        else:
            # {
            #     bodyDef->type = b2BodyType::b2_dynamicBody;
            body_def.type = b2.b2_dynamicBody
        # }
        #
        # //we set the position to the intial x,y coordinate
        # bodyDef->position.Set(entity["x"].asDouble(), entity["y"].asDouble());
        body_def.position = b2.b2Vec2(entity["x"], entity["y"])
        #
        # //we set physical properties of this body
        # //we don't want any rotation on the nodes (they would slide when moving)
        # bodyDef->linearDamping = 1.1f;
        body_def.linearDamping = 1.1
        #
        # //bodies have fixed rotation in iesor -- no move please k thx
        # bodyDef->fixedRotation = true;
        body_def.fixedRotation = True
        #
        # if(!entity["angle"].isNull())
        #     bodyDef->angle = entity["angle"].asDouble();
        if entity.get("angle", None) is not None:
            body_def.angle = entity["angle"]
        #
        # //we add our body to the world, and register it's identifying information
        # //that's useful for drawing
        # b2Body* body = this->addBodyToWorld(entity["bodyID"].asString(), bodyDef);
        body = self.add_body_to_world(entity["bodyID"], body_def)
        #
        # //we check if we have a radius (that makes us a circle entity)
        # if (!entity["radius"].isNull())
        if entity.get("radius", None) is not None:
            # {
            #     //this filter prevents circle nodes from colliding on our bodies
            #     b2Filter* filter = new b2Filter();
            filter = b2.b2Filter()
            #     filter->categoryBits = 0x0002;
            filter.categoryBits = 0x0002
            #     filter->maskBits = 0x0001;
            filter.maskBits = 0x0001
            #     filter->groupIndex = 0;
            filter.groupIndex = 0
            #
            #     //we create a fixture def for our new shape
            #     b2FixtureDef fixture;
            fixture = b2.b2FixtureDef()
            #
            #     //We need high density objects so they don't float!
            #     fixture.density = 25.0;
            fixture.density = 25.0
            #     //lots of friction please!
            #     fixture.friction = 1.0;
            fixture.friction = 1.0
            #     //less bouncing, more staying
            #     fixture.restitution = .1;
            fixture.restitution = 0.1
            #
            #     //we're a circle shape, make it so
            #     b2CircleShape node;
            node = b2.b2CircleShape()
            #
            #     //and we establish our radius
            #     node.m_radius = entity["radius"].asDouble();
            node.radius = entity["radius"]
            #
            #     //attach the shape to our fixtureDef
            #     fixture.shape = &node;
            fixture.shape = node
            #     fixture.filter = *filter;
            fixture.filter = filter
            #
            #     //add this fixture to our body (and track it in this class)
            #     this->addShapeToBody(body, &fixture);
            self.add_shape_to_body_fixture(body, fixture)
        # }
        # //now we're checking if we're a polygonal shape
        # else if (!entity["polyPoints"].isNull())
        elif entity.get("polyPoints", None) is not None:
            # {
            #     //we know how many points we need to add
            #     int pointCount = entity["polyPoints"].size();
            point_count = len(entity["polyPoints"])
            #
            #     //so we create an array to hold those points in a form Box2D undertstands
            #     b2Vec2* polyPoints = new b2Vec2[pointCount];
            #     //loop through our json points
            #     for (int j = 0; j < pointCount; j++)
            #     {
            #         //pull the object from our json array
            #         Json::Value point = entity["polyPoints"][j];
            #
            #         //convert to box2d coords
            #         b2Vec2 vec(point["x"].asDouble(), point["y"].asDouble());
            #
            #         //set our coordinate inside the array
            #         polyPoints[j] = vec;
            #     }
            poly_points = [
                b2.b2Vec2(point["x"], point["y"]) for point in entity["polyPoints"]
            ]
            #
            #     //create our polygon shape
            #     b2PolygonShape poly;
            poly = b2.b2PolygonShape()
            #
            #     //set our box2d points as the polygonal shape
            #     poly.Set(polyPoints, pointCount);
            poly.vertices = poly_points
            #
            #     //defalult density to 0, and add the sucker
            #     this->addShapeToBody(body, &poly, 0.0);
            self.add_shape_to_body(body, poly, 0.0)
        # }
        # else
        else:
            # {
            #     //we're a rectangle entity!
            #     b2PolygonShape poly;
            poly = b2.b2PolygonShape
            #
            #     //we've got a box, we simply put in the half height/ hafl width info for our polygon
            #     //it does the rest
            #     poly.SetAsBox(entity["halfWidth"].asDouble(),entity["halfHeight"].asDouble());
            poly.SetAsBox(entity["halfWidth"], entity["halfHeight"])
            #
            #     //defalult density to 0, and add the sucker
            #     this->addShapeToBody(body, &poly, 0.0);
            self.add_shape_to_body(body, poly, 0.0)
        # }
        #
        # //we've created our body and our shape, and it's all registered in the class
        # }
        return

    def create_shape_id(self) -> PhysicsId:
        # PhysicsID* IESoRWorld::createShapeID()
        # {
        #     //We just keep a static int of all the shapes for all identification
        #     int bcnt = staticBodyCount++;
        #
        #     //Track the shape string
        #     PhysicsID* shapeID = new PhysicsID(toString(bcnt));
        #
        #     //Save our shape to the list of shapes
        #     this->shapeList.push_back(shapeID);
        #
        #     return shapeID;
        # }
        global staticBodyCount
        staticBodyCount = staticBodyCount + 1
        shape_id = PhysicsId(str(staticBodyCount))
        # self.static_body_count += 1
        # shape_id = PhysicsId(str(self.static_body_count))
        self.shape_list.append(shape_id)
        return shape_id


def get_file_contents(filename: str) -> str:
    with open(filename) as f:
        return f.read()


def set_shape_id(fix: b2.b2Fixture, shape_id: PhysicsId):
    fix.userData = shape_id


def position_to_json_value(vec: b2.b2Vec2) -> dict:
    #
    # //converts a b2Vec point into a json object
    # Json::Value positionToJSONValue(b2Vec2 vec)
    # {
    #     Json::Value pos;
    #     pos["x"] = vec.x;
    #     pos["y"] = vec.y;
    #     return pos;
    # }
    return {"x": vec.x, "y": vec.y}


def list_of_points(points: list[b2.b2Vec2], length: int = 0) -> list:
    # //converts an array of points into a json array
    # Json::Value listOfPoints(b2Vec2* points, int length)
    # {
    #     Json::Value pArray(Json::arrayValue);
    #
    #     for(int i=0;i<length;i++)
    #     {
    #         pArray[i] = positionToJSONValue(points[i]);
    #     }
    #
    #     return pArray;
    # }

    # Pretty sure we don't need length...
    # Assuming it works.
    return [position_to_json_value(points[i]) for i in range(length)]


class SodaraceSimulator:
    def __init__(self, body: dict) -> None:
        self.world = IESoRWorld()
        self._morphology = self.world.load_body_into_world(body, self.world.canvas)

    @property
    def morphology(self) -> dict:
        return self._morphology

    def evaluate(self, timesteps: int) -> float:
        for _ in range(timesteps // 350):
            self.world.updateWorld(350)

        _ = self.morphology["startX"]
        end = min(
            [bone.joint.bodyA.position[0] for bone in self.world.bone_list]
            + [muscle.joint.bodyA.position[0] for muscle in self.world.muscle_list]
        )
        return abs(end + self.morphology["offsetX"])


# TODO: refactor into test
if __name__ == "__main__":
    import Box2D.Box2D as b2
    import simulator

    myWorld = simulator.IESoRWorld()
    with open("Physics/Data/jsBody142856.json", "r") as f:
        initial_dict = myWorld.load_body_into_world(json.load(f), myWorld.canvas)
    myWorld.updateWorld(10000.0)  # or however many ms
    start = initial_dict["startX"]
    end = min(
        [bone.joint.bodyA.position[0] for bone in myWorld.bone_list]
        + [muscle.joint.bodyA.position[0] for muscle in myWorld.muscle_list]
    )
    print(start, end + initial_dict["offsetX"])



from openelm.environments.sodaracer.simulator import IESoRWorld
from openelm.environments.sodaracer.walker import Walker
from openelm.environments.sodaracer.walker.CPPN_fixed import (
    make_walker as make_walker_cppn_fixed,
)
from openelm.environments.sodaracer.walker.CPPN_mutable import (
    make_walker as make_walker_cppn_mutable,
)
from openelm.environments.sodaracer.walker.radial import make_walker as make_walker_radial
from openelm.environments.sodaracer.walker.square import make_walker as make_walker_square


def test_sodaracer_eval():
    """Test that the Sodarace seeds give the correct evaluation."""
    square_walker: Walker = make_walker_square()
    radial_walker: Walker = make_walker_radial()
    cppn_fixed_walker: Walker = make_walker_cppn_fixed()
    cppn_mutable_walker: Walker = make_walker_cppn_mutable()

    square_world = IESoRWorld()
    radial_world = IESoRWorld()
    cppn_fixed_world = IESoRWorld()
    cppn_mutable_world = IESoRWorld()

    square_dict = square_world.load_body_into_world(square_walker.to_dict(),
                                                    square_world.canvas)
    assert square_dict == {
        'width': 413.2231404958677,
        'height': 240.0,
        'startX': 0.0,
        'offsetX': -106.61157024793386,
        'startY': -216.0,
        'offsetY': 171.0,
        'mass': 1136.0864722563908
    }

    def run_world(myWorld, initial_dict):
        [myWorld.updateWorld(350) for i in range(1000)]
        start = initial_dict['startX']
        # Could potentially just grab the bones instead, but if it's all
        # muscle then you'd need to grab a muscle.
        end = min(
            [bone.joint.bodyA.position.x for bone in myWorld.bone_list] +
            [muscle.joint.bodyA.position.x for muscle in myWorld.muscle_list]
        )
        return start, abs(end + initial_dict['offsetX'])

    square_start, square_end = run_world(square_world, square_dict)
    assert (square_start, square_end) == (0.0, 7.27755535535573)



from openelm.sandbox.server.environments.walker.walk_creator import walker_creator


def query_cppn(wc, xgrid, ygrid, scale, connect_func, amp_func, phase_func):
    """Create a grid of points and functionally connect them."""
    joints = {}
    for x in range(xgrid):
        for y in range(ygrid):
            joints[(x, y)] = wc.add_joint(x * scale, y * scale)
    for x1 in range(xgrid):
        for y1 in range(ygrid):
            for x2 in range(x1, xgrid):
                for y2 in range(y1, ygrid):
                    if x1 == y1 and x2 == y2:
                        continue
                    if connect_func(x1, y1, x2, y2):
                        amp = amp_func(x1, y1, x2, y2)
                        phase = phase_func(x1, y1, x2, y2)
                        wc.add_muscle(joints[(x1, y1)], joints[(x2, y2)], amp, phase)
    return joints


def make_walker():
    wc = walker_creator()

    def connect(x1, y1, x2, y2):
        if ((x1 - x2) ** 2 + (y1 - y2) ** 2) > 4.5:
            return False
        return True

    def amp(x1, y1, x2, y2):
        return max(abs(x1 - x2), abs(y1 - y2))

    def phase(x1, y1, x2, y2):
        return x1 if x1 % 2 == 1 else -x1

    _ = query_cppn(wc, 8, 3, 1.5, connect, amp, phase)

    return wc.get_walker()



import itertools

from openelm.utils.diff_eval import (
    DiffState,
    apply_diff,
    parse_line_info,
    replace_text,
    split_diff,
    verify_diff,
)


def test_diff():
    test = "<NME> test.py\n<BEF> asdafdsf\nasdlfjalskj\n  asldfajlk\n<MSG> asldkjf\n<DFF> asdkjjfajjsdfj\najföj"
    result = split_diff(test)
    assert result["name"] == "test.py"
    assert result["file"] == "asdafdsf\nasdlfjalskj\n  asldfajlk"
    assert result["message"] == "asldkjf"
    assert result["diff"] == "asdkjjfajjsdfj\najföj"

    test = "@@ -11,22 +33,45678 @@"
    result = parse_line_info(test)
    assert result == (11, 22, 33, 45678)

    # More serious test on an actual diff
    test = (
        "<NME> conf_template.lua\n<BEF> local _M = {}\n\nfunction _M:get_ngx_conf_template()\n    return [[\n# "
        "user www www;\npid tmp/{{LOR_ENV}}-nginx.pid;\n\n# This number should be at maxium the number of CPU on "
        "the server\nworker_processes 4;\n\nevents {\n    # Number of connections per worker\n    "
        "worker_connections 4096;\n}\n\nhttp {\n    sendfile on;\n    include ./mime.types;\n\n    {{"
        "LUA_PACKAGE_PATH}}\n    lua_code_cache on;\n\n    server {\n        # List port\n        listen {{"
        "PORT}};\n\n        # Access log\n        access_log logs/{{LOR_ENV}}-access.log;\n\n        # Error log\n "
        "       error_log logs/{{LOR_ENV}}-error.log;\n\n        # this variable is for view "
        "render（lua-resty-template)\n        set $template_root '';\n\n        location /static {\n            "
        "alias {{STATIC_FILE_DIRECTORY}}; #app/static;\n        }\n\n        # lor runtime\n        {{"
        "CONTENT_BY_LUA_FILE}}\n    }\n}\n]]\nend\n\nreturn _M\n\n<MSG> Merge pull request #100 from "
        "hanxi/patch-2\n\nfix a typo\n<DFF> @@ -30,7 +30,7 @@ http {\n         # Error log\n         error_log "
        "logs/{{LOR_ENV}}-error.log;\n \n-        # this variable is for view render（lua-resty-template)\n+        "
        "# this variable is for view render(lua-resty-template)\n         set $template_root '';\n \n         "
        "location /static {\n"
    )
    result = split_diff(test)
    # Because it is a valid diff, whichever method we apply the patch should result in the same.
    assert apply_diff(
        result["file"], result["diff"], use_line_number=False
    ) == apply_diff(result["file"], result["diff"], use_line_number=True)
    # Reject if pre-diff lines don't fully match
    assert (
        apply_diff(result["file"], result["diff"] + " ", use_line_number=True)
        == result["file"]
    )

    # Test ADDFILE
    test = (
        "<NME> test.py\n<BEF> ADDFILE\n<MSG> asldkjf\n<DFF> @@ -1,0 +0,0 @@\n"
        "+aaaaaaa\n+bbbbbbb"
    )
    result = split_diff(test)
    assert (
        apply_diff(result["file"], result["diff"], use_line_number=False)
        == "aaaaaaa\nbbbbbbb"
    )
    assert apply_diff(result["file"], result["diff"], use_line_number=True) == ""

    # Test the loose patching (still performs the patching until invalid diff lines)
    test = (
        "<NME> test.py\n<BEF> cccccc\n<MSG> asldkjf\n<DFF> @@ -1,1 +1,3 @@\n"
        "+aaaaaaa\n+bbbbbbb\n cccccc\ninvalid\n"
    )
    result = split_diff(test)
    assert (
        apply_diff(result["file"], result["diff"], use_line_number=False)
        == "aaaaaaa\nbbbbbbb\ncccccc"
    )
    # the first part -1,1 is consistent, so perform the patching
    assert (
        apply_diff(result["file"], result["diff"], use_line_number=True)
        == "aaaaaaa\nbbbbbbb\ncccccc"
    )

    # Test multiple @@ ... @@
    test = (
        "<NME> test.py\n<BEF> cccccc\ndddddd\n<MSG> asldkjf\n<DFF> @@ -1,1 +1,3 @@\n"
        "+aaaaaaa\n+bbbbbbb\n cccccc\n@@ -2,1 +4,1 @@\n-dddddd\n+eeeeee\n"
    )
    result = split_diff(test)
    assert (
        apply_diff(result["file"], result["diff"], use_line_number=False)
        == "aaaaaaa\nbbbbbbb\ncccccc\neeeeee"
    )
    assert (
        apply_diff(result["file"], result["diff"], use_line_number=True)
        == "aaaaaaa\nbbbbbbb\ncccccc\neeeeee"
    )
    # Test adding new line on the top, plus the default behavior of line range
    test = (
        "<NME> test.py\n<BEF> cccccc\ndddddd\n<MSG> asldkjf\n<DFF> @@ -0,0 +1,3 @@\n"
        "+aaaaaaa\n+bbbbbbb\n@@ -2 +4 @@\n-dddddd\n+eeeeee\n"
    )
    result = split_diff(test)
    assert apply_diff(result["file"], result["diff"], use_line_number=False) == "aaaaaaa\nbbbbbbb\ncccccc\neeeeee"
    assert apply_diff(result["file"], result["diff"], use_line_number=True) == "aaaaaaa\nbbbbbbb\ncccccc\neeeeee"


def test_replace_text():
    test = "abcdawasd\nfwgvwavwefadwwasd"
    before = "wasd"
    after = "1234"
    assert replace_text(test, before, after, 0) == ("abcda1234\nfwgvwavwefadwwasd", 9)

    test = "asdsdsadafwasd"
    assert replace_text(test, before, after, 0) == ("asdsdsadaf1234", 14)


def test_verify_diff():
    """
    Go over as many bad diff as possible to test the status code of `verify_diff`
    """

    # Valid diff
    test = (
        "<NME> test.py\n<BEF> cccccc\n<MSG> asldkjf\n<DFF> @@ -1,1 +1,3 @@\n"
        "+aaaaaaa\n+bbbbbbb\n cccccc\n"
    )
    assert verify_diff(test) == DiffState(0)
    # One can ignore the line range (default to 1)
    test = (
        "<NME> test.py\n<BEF> cccccc\n<MSG> asldkjf\n<DFF> @@ -1 +1,3 @@\n"
        "+aaaaaaa\n+bbbbbbb\n cccccc\n"
    )
    assert verify_diff(test) == DiffState(0)
    # Multiple valid
    test = (
        "<NME> test.py\n<BEF> cccccc\ndddddd\n<MSG> asldkjf\n<DFF> @@ -1,1 +1,3 @@\n"
        "+aaaaaaa\n+bbbbbbb\n cccccc\n@@ -2,1 +4,1 @@\n-dddddd\n+eeeeee\n"
    )
    assert verify_diff(test) == DiffState(0)

    # Test ADDFILE
    test = (
        "<NME> test.py\n<BEF> ADDFILE\n<MSG> asldkjf\n<DFF> @@ -1,1 +1,3 @@\n"
        "+aaaaaaa\n+bbbbbbb\n cccccc\n"
    )
    assert verify_diff(test) == DiffState(6)  # Wrong line numbers
    test = (
        "<NME> test.py\n<BEF> ADDFILE\n<MSG> asldkjf\n<DFF> @@ -0,0 +0,3 @@\n"
        "+aaaaaaa\n+bbbbbbb\n cccccc\n"
    )
    assert verify_diff(test) == DiffState(6)  # Contains lines not starting with "+"
    test = (
        "<NME> test.py\n<BEF> ADDFILE\n<MSG> asldkjf\n<DFF> @@ -0,0 +1,3 @@\n"
        "+aaaaaaa\n+bbbbbbb\n+cccccc\n"
    )
    assert verify_diff(test) == DiffState(0)

    test = (
        "<NME> test.py\n<BEF> cccccc\n<MSG> asldkjf\n<DFF> @@ -1,1 +1,3 @@\n"
        "+aaaaaaa\n+bbbbbbb\n ccccc\n"
    )
    assert verify_diff(test) == DiffState(1)
    for a, b, c, d in itertools.product([0, 2], repeat=4):
        for e in ["c", ""]:
            test = (
                f"<NME> test.py\n<BEF> cccccc\n<MSG> asldkjf\n<DFF> @@ -{a},{b} +{c},{d} @@\n"
                f"+aaaaaaa\n+bbbbbbb\n ccccc{e}\n"
            )
            assert verify_diff(test) == DiffState(2) if e == "c" else DiffState(3)
    test = (
        "<NME> test.py\n<BFE> cccccc\n<MSG> asldkjf\n<DFF> @@ -1,1 +1,4 @@\n"
        "+aaaaaaa\n+bbbbbbb\n cccccc\n"
    )
    assert verify_diff(test) == DiffState(4)  # Invalid format (BFE instead of BEF)
    test = (
        "<NME> test.py\n<BEF> cccccc\n<MSG> asldkjf\n<DFF> @@ -1,1 +1,3 @@\n"
        "+aaaaaaa\n+bbbbbbb\n*cccccc\n"
    )
    assert verify_diff(test) == DiffState(5)  # Invalid hunk (start with other char)
    test = (
        "<NME> test.py\n<BEF> cccccc\n<MSG> asldkjf\n<DFF> @@ +1,1 -1,4 @@\n"
        "+aaaaaaa\n+bbbbbbb\n cccccc\n"
    )
    assert verify_diff(test) == DiffState(
        6
    )  # Invalid format (wrong format in @@ ... @@)
    test = (
        "<NME> test.py\n<BEF> cccccc\n<MSG> asldkjf\n<DFF> @@ +1,1 -1,4 @@\n"
        "+aaaaaaa\n+bbbbbbb\n*cccccc\n"
    )
    assert verify_diff(test) == DiffState(7)  # Both


def selected_real_world_diffs():
    diff1 = '<NME> Num.hs\n<BEF> module Data.Propagator.Num where\n\nimport Control.Monad\nimport ' \
            'Control.Monad\nimport Control.Monad.ST\nimport Data.Propagator.Cell\nimport ' \
            'Data.Propagator.Class\nimport Data.Propagator.Supported\nimport Numeric.Natural\n  plus :: Cell s a -> ' \
            'Cell s a -> Cell s a -> ST s ()\n  plus x y z = do\n    lift2 (+) x y z\n    lift2 (-) z x y \n    lift2 ' \
            '(-) z y z\n\n  times :: Cell s a -> Cell s a -> Cell s a -> ST s ()\n  times = lift2 (*)\n\n  cabs :: Eq ' \
            'a => Cell s a -> Cell s a -> ST s ()\n  cabs x y = do\n    lift1 abs x y\n    watch y $ \\b -> when (b ' \
            '== 0) $ write x 0\n  cabs :: Cell s a -> Cell s a -> ST s ()\n  default cabs :: (Num a, Eq a) => Cell s ' \
            'a -> Cell s a -> ST s ()\n  cabs x y = do\n    lift1 abs x y\n    watch y $ \\b -> when (b == 0) $ write ' \
            'x 0\n\n  csignum :: Cell s a -> Cell s a -> ST s ()\n  default csignum :: (Num a, Eq a) => Cell s a -> ' \
            'Cell s a -> ST s ()\n  csignum x y = do\n    lift1 signum x y\n    watch y $ \\b -> when (b == 0) $ ' \
            'write x 0\n\ninstance PropagatedNum Integer where\n      else watch y $ \\ b -> when (b /= 0) $ write x ' \
            '0\n  cabs = unify\n\ninstance PropagatedNum Int \ninstance PropagatedNum Word where\n  cabs = unify\n\n  ' \
            '    if c == 0 then do\n        watch x $ \\ a -> when (a /= 0) $ write y 0\n        watch y $ \\ b -> ' \
            'when (b /= 0) $ write x 0\n      else do\n        watch x $ \\ a -> write y (c `div` a)\n        watch y ' \
            '$ \\ b -> write x (c `div` b)\n  cabs = unify\n\ninstance PropagatedNum (Supported Natural) where\n  ' \
            'ctimes x y z = do\n    lift2 (*) x y z\n    watch z $ \\c ->\n      when (c == 0) $ do\n        watch x ' \
            '$ \\ a -> when (a /= 0) $ write y 0\n        watch y $ \\ b -> when (b /= 0) $ write x 0\n  cabs = ' \
            'unify\n\ninstance PropagatedNum Int\ninstance PropagatedNum (Supported Int)\n\ninstance PropagatedNum ' \
            'Word where\n  cabs = unify\n\ninstance PropagatedNum (Supported Word) where\n  cabs = ' \
            'unify\n\nctimesFractional :: (Eq a, Fractional a) => Cell s a -> Cell s a -> Cell s a -> ST s (' \
            ')\nctimesFractional x y z = do\n  watch x $ \\a ->\n    if a == 0\n    then write z 0\n    else do\n     ' \
            ' with y $ \\b -> write z (a*b)\n      with z $ \\c -> write y (c/a) -- a /= 0 determined above\n  watch ' \
            'y $ \\b ->\n    if b == 0\n    then write z 0\n    else do\n      with x $ \\a -> write z (a*b)\n      ' \
            'with z $ \\c -> write x (c/b) -- b /= 0 determined above\n  watch z $ \\c -> do\n    with x $ \\a -> ' \
            'when (a /= 0) $ write y (c/a)\n    with y $ \\b -> when (b /= 0) $ write x (c/b)\n\ninstance ' \
            'PropagatedNum Rational where\n  ctimes = ctimesFractional\n\ninstance PropagatedNum (Supported Rational) ' \
            'where\n  ctimes = ctimesFractional\n\ninstance PropagatedNum Double where\n  ctimes = ' \
            'ctimesFractional\n\ninstance PropagatedNum (Supported Double) where\n  ctimes = ' \
            'ctimesFractional\n\ninstance PropagatedNum Float where\n  ctimes = ctimesFractional\n\ninstance ' \
            'PropagatedNum (Supported Float) where\n  ctimes = ctimesFractional\n\nclass PropagatedNum a => ' \
            'PropagatedFloating a where\n  cexp :: Cell s a -> Cell s a -> ST s ()\n  default cexp :: Floating a => ' \
            'Cell s a -> Cell s a -> ST s ()\n  cexp x y = do\n    lift1 exp x y\n    lift1 log y x\n\n  csqrt :: ' \
            'Cell s a -> Cell s a -> ST s ()\n  default csqrt :: Floating a => Cell s a -> Cell s a -> ST s ()\n  ' \
            'csqrt x y = do\n    lift1 sqrt x y\n    lift1 (\\a -> a*a) y x\n\n  csin :: Cell s a -> Cell s a -> ST s ' \
            '()\n  default csin :: (Floating a, Ord a) => Cell s a -> Cell s a -> ST s ()\n  csin x y = do\n    lift1 ' \
            'sin x y\n    watch y $ \\b -> do\n       unless (abs b <= 1) $ fail "output of sin not between -1 and ' \
            '1"\n       write x (asin b)\n\n  ccos :: Cell s a -> Cell s a -> ST s ()\n  default ccos :: (Floating a, ' \
            'Ord a) => Cell s a -> Cell s a -> ST s ()\n  ccos x y = do\n    lift1 cos x y\n    watch y $ \\b -> do\n ' \
            '      unless (abs b <= 1) $ fail "output of cos not between -1 and 1"\n       write x (acos b)\n\n  ctan ' \
            ':: Cell s a -> Cell s a -> ST s ()\n  default ctan :: (Floating a, Ord a) => Cell s a -> Cell s a -> ST ' \
            's ()\n  ctan x y = do\n    lift1 tan x y\n    watch y $ \\b -> do\n      unless (abs b <= pi/2) $ fail ' \
            '"output of tan not between -pi/2 and pi/2"\n      write x (atan b)\n\n  csinh :: Cell s a -> Cell s a -> ' \
            'ST s ()\n  default csinh :: (Floating a, Ord a) => Cell s a -> Cell s a -> ST s ()\n  csinh x y = do\n   ' \
            ' lift1 sinh x y\n    lift1 asinh y x\n\n  ccosh :: Cell s a -> Cell s a -> ST s ()\n  default ccosh :: (' \
            'Floating a, Ord a) => Cell s a -> Cell s a -> ST s ()\n  ccosh x y = do\n    lift1 cosh x y\n    watch y ' \
            '$ \\b -> do\n      unless (b >= 1) $ fail "output of cosh not >= 1"\n      lift1 acosh y x\n\n  ctanh :: ' \
            'Cell s a -> Cell s a -> ST s ()\n  default ctanh :: (Floating a, Ord a) => Cell s a -> Cell s a -> ST s ' \
            '()\n  ctanh x y = do\n    lift1 tanh x y\n    watch y $ \\b -> do\n      unless (abs b <= 1) $ fail ' \
            '"output of tanh not between -1 and 1"\n      write x (tanh b)\n\ninstance PropagatedFloating ' \
            'Float\ninstance PropagatedFloating (Supported Float)\ninstance PropagatedFloating Double\ninstance ' \
            'PropagatedFloating (Supported Double)\n\n-- Interval arithmetic\n\nclass (Floating a, Ord a) => ' \
            'PropagatedInterval a where\n  infinity :: a\n\n\ninstance PropagatedInterval Double where\n  infinity = ' \
            '1/0\n\ninstance PropagatedInterval (Supported Double) where\n  infinity = 1/0\n\ninstance ' \
            'PropagatedInterval Float where\n  infinity = 1/0\n\ninstance PropagatedInterval (Supported Float) ' \
            'where\n  infinity = 1/0\n\ninstance PropagatedInterval a => PropagatedNum (Interval a) where\n  ctimes = ' \
            'ctimesFractional\n\n  cabs x y = do\n    write y (0...infinity)\n    lift1 abs x y\n    -- todo: use ' \
            'symmetric_positive\n    watch y $ \\case\n      I _ b -> write x (-b...b)\n      Empty -> write x ' \
            'Empty\n\n  csignum x y = do\n    write y (-1...1)\n    lift1 signum x y\n    watch y $ \\case\n      I a ' \
            'b | a < 1 && b > -1 -> write x $ I (if a <= -1 then -infinity else 0) (if b >= 1 then infinity else 0)\n ' \
            '     _ -> write x Empty\n\nsymmetric_positive :: (Num a, Ord a) => (Interval a -> Interval a) -> Cell s ' \
            '(Interval a) -> Cell s (Interval a) -> ST s ()\nsymmetric_positive f x y = do\n  watch y $ \\case\n    ' \
            'Empty -> write x Empty -- if the result is empty then the input is empty\n    I a a\' -> do\n      when ' \
            '(a\' <= 0) $ with x $ \\c -> write y (- f c)\n      when (a >= 0)  $ with x $ \\c -> write y (f c)\n  ' \
            'lift1 (\\c -> let d = f c in hull (-d) d) x y\n\n-- x = f y + p*n\n-- n = (x - f y)/p, ' \
            'n is an integer\nperiodic :: RealFrac a => Interval a -> (Interval a -> Interval a) -> Cell s (Interval ' \
            'a) -> Cell s (Interval a) -> ST s ()\nperiodic p f x y = do\n  watch2 x y $ \\a b -> let c = f b in case ' \
            '(a - c) / p of\n    Empty -> write x Empty\n    I l h -> write x (c + p*(fromIntegral (ceiling l :: ' \
            'Integer)...fromIntegral (floor h :: Integer)))\n\ninstance (PropagatedInterval a, RealFloat a) => ' \
            'PropagatedFloating (Interval a) where\n  cexp x y = do\n    write y (0...infinity)\n    lift1 exp x y\n  ' \
            '  lift1 log y x\n\n  csqrt x y = do\n    write x (0...infinity)\n    lift1 (\\b -> b*b) y x\n    ' \
            'symmetric_positive sqrt x y\n\n  csin x y = do\n    write y (-1...1)\n    lift1 sin x y\n    periodic (' \
            '2*pi) asin y x\n\n  ccos x y = do\n    write y (-1...1)\n    lift1 cos x y\n    periodic (2*pi) acos y ' \
            'x\n\n  ctan x y = do\n    write y (-pi/2...pi/2)\n    lift1 tan x y\n    periodic pi atan y x\n\n  csinh ' \
            'x y = do\n    lift1 sinh x y\n    lift1 asinh y x\n\n  ccosh x y = do\n    write y (1...infinity)\n    ' \
            'lift1 cosh x y\n    symmetric_positive acosh x y\n\n  ctanh x y = do\n    write y (-1...1)\n    lift1 ' \
            'tanh x y\n    lift1 atanh y x\n\n<MSG> PropagatedNum rational\n\n<DFF> @@ -1,3 +1,6 @@\n+{-# LANGUAGE ' \
            'DefaultSignatures #-}\n+{-# LANGUAGE FlexibleInstances #-}\n+{-# LANGUAGE TypeSynonymInstances #-}\n ' \
            'module Data.Propagator.Num where\n \n import Control.Monad\n@@ -10,13 +13,14 @@ class (Propagated a, ' \
            'Num a) => PropagatedNum a where\n   plus :: Cell s a -> Cell s a -> Cell s a -> ST s ()\n   plus x y z = ' \
            'do\n     lift2 (+) x y z\n-    lift2 (-) z x y \n+    lift2 (-) z x y\n     lift2 (-) z y z\n \n   times ' \
            ':: Cell s a -> Cell s a -> Cell s a -> ST s ()\n   times = lift2 (*)\n \n-  cabs :: Eq a => Cell s a -> ' \
            'Cell s a -> ST s ()\n+  cabs :: Cell s a -> Cell s a -> ST s ()\n+  default cabs :: Eq a => Cell s a -> ' \
            'Cell s a -> ST s ()\n   cabs x y = do\n     lift1 abs x y\n     watch y $ \\b -> when (b == 0) $ write x ' \
            '0\n@@ -36,7 +40,25 @@ instance PropagatedNum Natural where\n       else watch y $ \\ b -> when (b /= 0) ' \
            '$ write x 0\n   cabs = unify\n \n-instance PropagatedNum Int \n+instance PropagatedNum Int\n+\n instance ' \
            'PropagatedNum Word where\n   cabs = unify\n \n+instance PropagatedNum Rational where\n+  times x y z = ' \
            'do\n+    watch x $ \\a ->\n+      if a == 0\n+      then write z 0\n+      else do\n+        with y $ ' \
            '\\b -> write z (a*b)\n+        with z $ \\c -> write y (c/a) -- a /= 0 determined above\n+    watch y $ ' \
            '\\b ->\n+      if b == 0\n+      then write z 0\n+      else do\n+        with x $ \\a -> write z (' \
            'a*b)\n+        with z $ \\c -> write x (c/b) -- b /= 0 determined above\n+    watch z $ \\c -> do\n+     ' \
            ' with x $ \\a -> when (a /= 0) $ write y (c/a)\n+      with y $ \\b -> when (b /= 0) $ write x (c/b)\n '

    assert verify_diff(diff1).value == 0



from openelm.environments.environments import (
    BaseEnvironment,
    FunctionOptim,
    ImageOptim,
    MatchString,
)
from openelm.map_elites import MAPElites
from openelm.environments import image_init_args


def test_string_matching():
    env: BaseEnvironment = MatchString(target="MAPElites")
    elites = MAPElites(env, n_bins=3, history_length=10)
    assert elites.search(initsteps=10_000, totalsteps=100_000) == "MAPElites"

    elites.plot()


def test_function_optim():
    env: BaseEnvironment = FunctionOptim(ndim=2)
    elites = MAPElites(env, n_bins=128, history_length=10)
    assert elites.search(initsteps=5_000, totalsteps=50_000, atol=0) == 1.0

    elites.plot()


def test_image_optim():
    env = ImageOptim(**image_init_args)
    elites = MAPElites(env, n_bins=2, history_length=10)

    print("Best image", elites.search(initsteps=5, totalsteps=10))



from collections import defaultdict
from typing import Optional

import numpy as np
from tqdm import trange

from openelm.environments import BaseEnvironment

Phenotype = Optional[np.ndarray]
Mapindex = Optional[tuple]


class Map:
    """
    Class to represent a map of any dimensionality, backed by a numpy array.

    This class is necessary to handle the circular buffer for the history dimension.
    """
    def __init__(
        self,
        dims: tuple,
        fill_value: float,
        dtype: type = np.float32,
        history_length: int = 1,
    ):
        """
        Class to represent a map of any dimensionality, backed by a numpy array.

        This class is a wrapper around a numpy array that handles the circular
        buffer for the history dimension. We use it for the array of fitnesses,
        genomes, and to track whether a niche in the map has been explored or not.

        Args:
            dims (tuple): Tuple of ints representing the dimensions of the map.
            fill_value (float): Fill value to initialize the array.
            dtype (type, optional): Type to pass to the numpy array to initialize
                it. For example, we initialize the map of genomes with type `object`.
                Defaults to np.float32.
            history_length (int, optional): Length of history to store for each
                niche (cell) in the map. This acts as a circular buffer, so after
                storing `history_length` items, the buffer starts overwriting the
                oldest items. Defaults to 1.
        """
        self.history_length: int = history_length
        self.dims: tuple = dims
        if self.history_length == 1:
            self.array: np.ndarray = np.full(dims, fill_value, dtype=dtype)
        else:
            # Set starting top of buffer to 0 (% operator)
            self.top = np.full(dims, self.history_length - 1, dtype=int)
            self.array = np.full((history_length,) + dims, fill_value, dtype=dtype)
        self.empty = True

    def __getitem__(self, map_ix):
        """If history length > 1, the history dim is an n-dim circular buffer."""
        if self.history_length == 1:
            return self.array[map_ix]
        else:
            return self.array[(self.top[map_ix], *map_ix)]

    def __setitem__(self, map_ix, value):
        self.empty = False
        if self.history_length == 1:
            self.array[map_ix] = value
        else:
            top_val = self.top[map_ix]
            top_val = (top_val + 1) % self.history_length
            self.top[map_ix] = top_val
            self.array[(self.top[map_ix], *map_ix)] = value

    @property
    def shape(self) -> tuple:
        """Wrapper around the shape of the numpy array."""
        return self.array.shape

    @property
    def map_size(self) -> int:
        """Returns the product of the dimension sizes, not including history."""
        if self.history_length == 1:
            return self.array.size
        else:
            return self.array[0].size


class MAPElites:
    """
    Class implementing MAP-Elites, a quality-diversity algorithm.

    MAP-Elites creates a map of high perfoming solutions at each point in a
    discretized behavior space. First, the algorithm generates some initial random
    solutions, and evaluates them in the environment. Then, it  repeatedly mutates
    the solutions in the map, and places the mutated solutions in the map if they
    outperform the solutions already in their niche.
    """
    def __init__(
        self, env, n_bins: int, history_length: int, save_history: bool = False
    ):
        """
        Class implementing MAP-Elites, a quality-diversity algorithm.

        Args:
            env (BaseEnvironment): The environment to evaluate solutions in. This
            should be a subclass of `BaseEnvironment`, and should implement
            methods to generate random solutions, mutate existing solutions,
            and evaluate solutions for their fitness in the environment.
            n_bins (int): Number of bins to partition the behavior space into.
            history_length (int): Length of history to store for each niche (cell)
            in the map. This acts as a circular buffer, so after storing
            `history_length` items, the buffer starts overwriting the oldest
            items.
            save_history (bool, optional): Whether to save the history of all
            generated solutions, even if they are not inserted into the map.
            Defaults to False.
        """
        self.env: BaseEnvironment = env
        self.n_bins = n_bins
        self.history_length = history_length
        self.save_history = save_history
        # self.history will be set/reset each time when calling `.search(...)`
        self.history: dict = defaultdict(list)

        # discretization of space
        self.bins = np.linspace(*env.behavior_space, n_bins + 1)[1:-1].T  # type: ignore
        # perfomance of niches
        self.fitnesses: Map = Map(
            dims=(n_bins,) * env.behavior_ndim,
            fill_value=-np.inf,
            dtype=float,
            history_length=history_length,
        )
        # niches' sources
        self.genomes: Map = Map(
            dims=self.fitnesses.dims,
            fill_value=0.0,
            dtype=object,
            history_length=history_length,
        )
        # index over explored niches to select from
        self.nonzero: Map = Map(dims=self.fitnesses.dims, fill_value=False, dtype=bool)
        # bad mutations that ended up with invalid output.
        self.recycled = [None] * 1000
        self.recycled_count = 0

        print(f"MAP of size: {self.fitnesses.dims} = {self.fitnesses.map_size}")

    def to_mapindex(self, b: Phenotype) -> Mapindex:
        """Converts a phenotype (position in behaviour space) to a map index."""
        return (
            None
            if b is None
            else tuple(np.digitize(x, bins) for x, bins in zip(b, self.bins))
        )

    def random_selection(self) -> Mapindex:
        """Randomly select a niche (cell) in the map that has been explored."""
        ix = np.random.choice(np.flatnonzero(self.nonzero.array))
        return np.unravel_index(ix, self.nonzero.dims)

    def search(self, initsteps: int, totalsteps: int, atol=1) -> str:
        """
        Run the MAP-Elites search algorithm.

        Args:
            initsteps (int): Number of initial random solutions to generate.
            totalsteps (int): Total number of steps to run the algorithm for,
                including initial steps.
            atol (int, optional): Tolerance for how close the best performing
                solution has to be to the maximum possible fitness before the
                search stops early. Defaults to 1.

        Returns:
            str: A string representation of the best perfoming solution. The
                best performing solution object can be accessed via the
                `current_max_genome` class attribute.
        """
        tbar = trange(int(totalsteps))
        max_fitness = -np.inf
        max_genome = None
        if self.save_history:
            self.history = defaultdict(list)

        for n_steps in tbar:
            if n_steps < initsteps or self.genomes.empty:
                # Initialise by generating initsteps random solutions.
                # If map is still empty: force to do generation instead of mutation.
                new_individuals = self.env.random()
            else:
                # Randomly select an elite from the map.
                map_ix = self.random_selection()
                selected_elite = self.genomes[map_ix]
                # Mutate the elite.
                new_individuals = self.env.mutate(selected_elite)

            # `new_individuals` is a list of generation/mutation. We put them into the behavior space one-by-one.
            for individual in new_individuals:
                map_ix = self.to_mapindex(self.env.to_behavior_space(individual))
                # if the return is None, the individual is invalid and is thrown into the recycle bin.
                if map_ix is None:
                    self.recycled[self.recycled_count % len(self.recycled)] = individual
                    self.recycled_count += 1
                    continue

                if self.save_history:
                    self.history[map_ix].append(individual)
                self.nonzero[map_ix] = True

                fitness = self.env.fitness(individual)
                # If new fitness greater than old fitness in niche, replace.
                if fitness > self.fitnesses[map_ix]:
                    self.fitnesses[map_ix] = fitness
                    self.genomes[map_ix] = individual
                # If new fitness is the highest so far, update the tracker.
                if fitness > max_fitness:
                    max_fitness = fitness
                    max_genome = individual

                    tbar.set_description(f"{max_fitness=:.4f}")
                # If best fitness is within atol of the maximum possible fitness, stop.
                if np.isclose(max_fitness, self.env.max_fitness, atol=atol):
                    break

        self.current_max_genome = max_genome
        return str(max_genome)

    def plot(self):
        import matplotlib
        from matplotlib import pyplot

        matplotlib.rcParams["font.family"] = "Futura"
        matplotlib.rcParams["figure.dpi"] = 100
        matplotlib.style.use("ggplot")

        ix = tuple(np.zeros(self.fitnesses.array.ndim - 2, int))
        print(ix)
        map2d = self.fitnesses[ix]
        print(f"{map2d.shape=}")

        pyplot.pcolor(map2d, cmap="inferno")
        pyplot.show()

    def niches_filled(self):
        """Get the number of niches that have been explored in the map."""
        return np.count_nonzero(self.nonzero.array)

    def maximum_fitness(self):
        """Get the maximum fitness value in the map."""
        return self.fitnesses.array.max()

    def quality_diversity_score(self):
        """
        Get the quality-diversity score of the map.

        The quality-diversity score is the sum of the performance of all solutions
        in the map.
        """
        return self.fitnesses.array[np.isfinite(self.fitnesses.array)].sum()



# coding=utf-8
# Copyright 2021 The EleutherAI and HuggingFace Teams. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Modified forward-pass implementation based on https://github.com/huggingface/transformers/blob/main/src/transformers/models/gptj/modeling_gptj.py

from typing import Tuple

import torch
import torch.utils.checkpoint
from torch import nn
from torch.nn import CrossEntropyLoss
from transformers.activations import ACT2FN
from transformers.modeling_outputs import (
    BaseModelOutputWithPast,
    CausalLMOutputWithPast,
)
from transformers.modeling_utils import PreTrainedModel
from transformers.utils import logging
from transformers.utils.model_parallel_utils import assert_device_map, get_device_map

from .configuration_codegen import CodeGenConfig

logger = logging.get_logger(__name__)


def fixed_pos_embedding(x, seq_dim=1, seq_len=None):
    dim = x.shape[-1]
    if seq_len is None:
        seq_len = x.shape[seq_dim]
    inv_freq = 1.0 / (10000 ** (torch.arange(0, dim, 2) / dim))
    # original
    # sinusoid_inp = torch.einsum("i , j -> i j", torch.arange(seq_len), inv_freq).to(x.device).float()
    # QHD fix onnx error by https://github.com/microsoft/onnxruntime/discussions/10121#discussioncomment-1987845
    sinusoid_inp = (
        torch.einsum("i , j -> i j", torch.arange(seq_len).float(), inv_freq)
        .to(x.device)
        .float()
    )
    return torch.sin(sinusoid_inp), torch.cos(sinusoid_inp)


def rotate_every_two(x):
    x1 = x[:, :, :, ::2]
    x2 = x[:, :, :, 1::2]
    x = torch.stack((-x2, x1), axis=-1)
    return x.flatten(-2)  # in einsum notation: rearrange(x, '... d j -> ... (d j)')


def apply_rotary_pos_emb(x, sincos, offset=0):
    sin, cos = map(
        lambda t: t[None, offset : x.shape[1] + offset, None, :].repeat_interleave(
            2, 3
        ),
        sincos,
    )
    # einsum notation for lambda t: repeat(t[offset:x.shape[1]+offset,:], "n d -> () n () (d j)", j=2)
    return (x * cos) + (rotate_every_two(x) * sin)


class CodeGenAttention(nn.Module):
    def __init__(self, config):
        super().__init__()

        max_positions = config.max_position_embeddings
        self.register_buffer(
            "bias",
            torch.tril(
                torch.ones((max_positions, max_positions), dtype=torch.bool)
            ).view(1, 1, max_positions, max_positions),
        )
        self.register_buffer("masked_bias", torch.tensor(-1e9))

        self.attn_dropout = nn.Dropout(config.attn_pdrop)
        self.resid_dropout = nn.Dropout(config.resid_pdrop)

        self.embed_dim = config.hidden_size
        self.num_attention_heads = config.num_attention_heads
        self.head_dim = self.embed_dim // self.num_attention_heads
        if self.head_dim * self.num_attention_heads != self.embed_dim:
            raise ValueError(
                f"embed_dim must be divisible by num_attention_heads (got `embed_dim`: {self.embed_dim} and `num_attention_heads`: {self.num_attention_heads})."
            )
        self.scale_attn = torch.sqrt(
            torch.tensor(self.head_dim, dtype=torch.float32)
        ).to(torch.get_default_dtype())
        self.qkv_proj = nn.Linear(self.embed_dim, self.embed_dim * 3, bias=False)

        self.out_proj = nn.Linear(self.embed_dim, self.embed_dim, bias=False)
        self.rotary_dim = None
        if config.rotary_dim is not None:
            self.rotary_dim = config.rotary_dim

    def _split_heads(self, x, n_head, dim_head, mp_num):
        reshaped = x.reshape(x.shape[:-1] + (n_head // mp_num, dim_head))
        reshaped = reshaped.reshape(x.shape[:-2] + (-1,) + reshaped.shape[-1:])
        return reshaped

    def _merge_heads(self, tensor, num_attention_heads, attn_head_size):
        """
        Merges attn_head_size dim and num_attn_heads dim into n_ctx
        """
        if len(tensor.shape) == 5:
            tensor = tensor.permute(0, 1, 3, 2, 4).contiguous()
        elif len(tensor.shape) == 4:
            tensor = tensor.permute(0, 2, 1, 3).contiguous()
        else:
            raise ValueError(
                f"Input tensor rank should be one of [4, 5], but is: {len(tensor.shape)}"
            )
        new_shape = tensor.size()[:-2] + (num_attention_heads * attn_head_size,)
        return tensor.view(new_shape)

    def _attn(
        self,
        query,
        key,
        value,
        attention_mask=None,
        head_mask=None,
    ):

        # compute causal mask from causal mask buffer
        query_length, key_length = query.size(-2), key.size(-2)
        causal_mask = self.bias[
            :, :, key_length - query_length : key_length, :key_length
        ].to(torch.bool)

        # Keep the attention weights computation in fp32 to avoid overflow issues
        query = query.to(torch.float32)
        key = key.to(torch.float32)

        attn_weights = torch.matmul(query, key.transpose(-1, -2))

        attn_weights = attn_weights / self.scale_attn
        attn_weights = torch.where(
            causal_mask, attn_weights, self.masked_bias.to(attn_weights.dtype)
        )

        if attention_mask is not None:
            # Apply the attention mask
            attn_weights = attn_weights + attention_mask

        attn_weights = nn.Softmax(dim=-1)(attn_weights)
        attn_weights = attn_weights.to(value.dtype)
        attn_weights = self.attn_dropout(attn_weights)

        # Mask heads if we want to
        if head_mask is not None:
            attn_weights = attn_weights * head_mask

        attn_output = torch.matmul(attn_weights, value)

        return attn_output, attn_weights

    def forward(
        self,
        hidden_states,
        attention_mask=None,
        layer_past=None,
        head_mask=None,
        use_cache=False,
        output_attentions=False,
    ):

        qkv = self.qkv_proj(hidden_states)
        # TODO(enijkamp): factor out number of logical TPU-v4 cores or make forward pass agnostic
        mp_num = 4
        qkv_split = qkv.reshape(qkv.shape[:-1] + (mp_num, -1))

        local_dim = self.head_dim * self.num_attention_heads // mp_num
        query, value, key = torch.split(qkv_split, local_dim, dim=-1)
        query = self._split_heads(
            query, self.num_attention_heads, self.head_dim, mp_num=mp_num
        )
        key = self._split_heads(
            key, self.num_attention_heads, self.head_dim, mp_num=mp_num
        )

        value = self._split_heads(
            value, self.num_attention_heads, self.head_dim, mp_num=mp_num
        )
        value = value.permute(0, 2, 1, 3)

        seq_len = key.shape[1]
        offset = 0

        if layer_past is not None:
            offset = layer_past[0].shape[-2]
            seq_len += offset

        if self.rotary_dim is not None:
            k_rot = key[:, :, :, : self.rotary_dim]
            k_pass = key[:, :, :, self.rotary_dim :]

            q_rot = query[:, :, :, : self.rotary_dim]
            q_pass = query[:, :, :, self.rotary_dim :]

            sincos = fixed_pos_embedding(k_rot, 1, seq_len=seq_len)
            k_rot = apply_rotary_pos_emb(k_rot, sincos, offset=offset)
            q_rot = apply_rotary_pos_emb(q_rot, sincos, offset=offset)

            key = torch.cat([k_rot, k_pass], dim=-1)
            query = torch.cat([q_rot, q_pass], dim=-1)
        else:
            sincos = fixed_pos_embedding(key, 1, seq_len=seq_len)
            key = apply_rotary_pos_emb(key, sincos, offset=offset)
            query = apply_rotary_pos_emb(query, sincos, offset=offset)

        key = key.permute(0, 2, 1, 3)
        query = query.permute(0, 2, 1, 3)

        if layer_past is not None:
            past_key = layer_past[0]
            past_value = layer_past[1]
            key = torch.cat((past_key, key), dim=-2)
            value = torch.cat((past_value, value), dim=-2)

        if use_cache is True:
            present = (key, value)
        else:
            present = None

        # compute self-attention: V x Softmax(QK^T)
        attn_output, attn_weights = self._attn(
            query, key, value, attention_mask, head_mask
        )

        attn_output = self._merge_heads(
            attn_output, self.num_attention_heads, self.head_dim
        )

        attn_output = self.out_proj(attn_output)
        attn_output = self.resid_dropout(attn_output)

        outputs = (attn_output, present)
        if output_attentions:
            outputs += (attn_weights,)

        return outputs  # a, present, (attentions)


class CodeGenMLP(nn.Module):
    def __init__(
        self, intermediate_size, config
    ):  # in MLP: intermediate_size= 4 * embed_dim
        super().__init__()
        embed_dim = config.n_embd

        self.fc_in = nn.Linear(embed_dim, intermediate_size)
        self.fc_out = nn.Linear(intermediate_size, embed_dim)

        self.act = ACT2FN[config.activation_function]
        self.dropout = nn.Dropout(config.resid_pdrop)

    def forward(self, hidden_states):
        hidden_states = self.fc_in(hidden_states)
        hidden_states = self.act(hidden_states)
        hidden_states = self.fc_out(hidden_states)
        hidden_states = self.dropout(hidden_states)
        return hidden_states


class CodeGenBlock(nn.Module):
    def __init__(self, config):
        super().__init__()
        inner_dim = config.n_inner if config.n_inner is not None else 4 * config.n_embd
        self.ln_1 = nn.LayerNorm(config.n_embd, eps=config.layer_norm_epsilon)
        self.attn = CodeGenAttention(config)
        self.mlp = CodeGenMLP(inner_dim, config)

    def forward(
        self,
        hidden_states,
        layer_past=None,
        attention_mask=None,
        head_mask=None,
        use_cache=False,
        output_attentions=False,
    ):
        residual = hidden_states
        hidden_states = self.ln_1(hidden_states)
        attn_outputs = self.attn(
            hidden_states,
            layer_past=layer_past,
            attention_mask=attention_mask,
            head_mask=head_mask,
            use_cache=use_cache,
            output_attentions=output_attentions,
        )
        attn_output = attn_outputs[0]  # output_attn: a, present, (attentions)
        outputs = attn_outputs[1:]

        feed_forward_hidden_states = self.mlp(hidden_states)
        hidden_states = attn_output + feed_forward_hidden_states + residual

        if use_cache:
            outputs = (hidden_states,) + outputs
        else:
            outputs = (hidden_states,) + outputs[1:]

        return outputs  # hidden_states, present, (attentions)


class CodeGenPreTrainedModel(PreTrainedModel):
    """
    An abstract class to handle weights initialization and a simple interface for downloading and loading pretrained
    models.
    """

    config_class = CodeGenConfig
    base_model_prefix = "transformer"
    is_parallelizable = True

    def __init__(self, *inputs, **kwargs):
        super().__init__(*inputs, **kwargs)

    def _init_weights(self, module):
        """Initialize the weights."""
        if isinstance(module, (nn.Linear,)):
            # Slightly different from Mesh Transformer JAX which uses truncated_normal for initialization
            # cf https://github.com/pytorch/pytorch/pull/5617
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.Embedding):
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
            if module.padding_idx is not None:
                module.weight.data[module.padding_idx].zero_()
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)


class CodeGenModel(CodeGenPreTrainedModel):
    def __init__(self, config):
        super().__init__(config)

        self.embed_dim = config.n_embd
        self.vocab_size = config.vocab_size
        self.wte = nn.Embedding(config.vocab_size, self.embed_dim)
        self.drop = nn.Dropout(config.embd_pdrop)
        self.h = nn.ModuleList([CodeGenBlock(config) for _ in range(config.n_layer)])
        self.ln_f = nn.LayerNorm(self.embed_dim, eps=config.layer_norm_epsilon)
        self.rotary_dim = min(
            config.rotary_dim, config.n_ctx // config.num_attention_heads
        )
        self.init_weights()

        # Model parallel
        self.model_parallel = False
        self.device_map = None

    def parallelize(self, device_map=None):
        # Check validity of device_map
        self.device_map = (
            get_device_map(len(self.h), range(torch.cuda.device_count()))
            if device_map is None
            else device_map
        )
        assert_device_map(self.device_map, len(self.h))
        self.model_parallel = True
        self.first_device = (
            "cpu"
            if "cpu" in self.device_map.keys()
            else "cuda:" + str(min(self.device_map.keys()))
        )
        self.last_device = "cuda:" + str(max(self.device_map.keys()))
        self.wte = self.wte.to(self.first_device)
        # Load onto devices
        for k, v in self.device_map.items():
            for block in v:
                cuda_device = "cuda:" + str(k)
                self.h[block] = self.h[block].to(cuda_device)
        # ln_f to last
        self.ln_f = self.ln_f.to(self.last_device)

    def deparallelize(self):
        self.model_parallel = False
        self.device_map = None
        self.first_device = "cpu"
        self.last_device = "cpu"
        self.wte = self.wte.to("cpu")
        for index in range(len(self.h)):
            self.h[index] = self.h[index].to("cpu")
        self.ln_f = self.ln_f.to("cpu")
        torch.cuda.empty_cache()

    def get_input_embeddings(self):
        return self.wte

    def set_input_embeddings(self, new_embeddings):
        self.wte = new_embeddings

    def forward(
        self,
        input_ids=None,
        past_key_values=None,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        head_mask=None,
        inputs_embeds=None,
        use_cache=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
    ):
        output_attentions = (
            output_attentions
            if output_attentions is not None
            else self.config.output_attentions
        )
        output_hidden_states = (
            output_hidden_states
            if output_hidden_states is not None
            else self.config.output_hidden_states
        )
        use_cache = use_cache if use_cache is not None else self.config.use_cache
        return_dict = (
            return_dict if return_dict is not None else self.config.use_return_dict
        )

        if input_ids is not None and inputs_embeds is not None:
            raise ValueError(
                "You cannot specify both input_ids and inputs_embeds at the same time"
            )
        elif input_ids is not None:
            input_shape = input_ids.size()
            input_ids = input_ids.view(-1, input_shape[-1])
            batch_size = input_ids.shape[0]
        elif inputs_embeds is not None:
            input_shape = inputs_embeds.size()[:-1]
            batch_size = inputs_embeds.shape[0]
        else:
            raise ValueError("You have to specify either input_ids or inputs_embeds")

        device = input_ids.device if input_ids is not None else inputs_embeds.device

        if token_type_ids is not None:
            token_type_ids = token_type_ids.view(-1, input_shape[-1])

        if position_ids is not None:
            position_ids = position_ids.view(-1, input_shape[-1])

        if past_key_values is None:
            past_length = 0
            past_key_values = tuple([None] * len(self.h))
        else:
            past_length = past_key_values[0][0].size(-2)

        if position_ids is None:
            position_ids = torch.arange(
                past_length,
                input_shape[-1] + past_length,
                dtype=torch.long,
                device=device,
            )
            position_ids = position_ids.unsqueeze(0).view(-1, input_shape[-1])

        # Attention mask.
        if attention_mask is not None:
            assert batch_size > 0, "batch_size has to be defined and > 0"
            attention_mask = attention_mask.view(batch_size, -1)
            # We create a 3D attention mask from a 2D tensor mask.
            # Sizes are [batch_size, 1, 1, to_seq_length]
            # So we can broadcast to [batch_size, num_heads, from_seq_length, to_seq_length]
            # this attention mask is more simple than the triangular masking of causal attention
            # used in OpenAI GPT, we just need to prepare the broadcast dimension here.
            attention_mask = attention_mask[:, None, None, :]

            # Since attention_mask is 1.0 for positions we want to attend and 0.0 for
            # masked positions, this operation will create a tensor which is 0.0 for
            # positions we want to attend and -10000.0 for masked positions.
            # Since we are adding it to the raw scores before the softmax, this is
            # effectively the same as removing these entirely.
            attention_mask = attention_mask.to(dtype=self.dtype)  # fp16 compatibility
            attention_mask = (1.0 - attention_mask) * -10000.0

        # Prepare head mask if needed
        # 1.0 in head_mask indicate we keep the head
        # attention_probs has shape bsz x num_attention_heads x N x N
        # head_mask has shape n_layer x batch x num_attention_heads x N x N
        head_mask = self.get_head_mask(head_mask, self.config.n_layer)

        if inputs_embeds is None:
            inputs_embeds = self.wte(input_ids)

        hidden_states = inputs_embeds

        if token_type_ids is not None:
            token_type_embeds = self.wte(token_type_ids)
            hidden_states = hidden_states + token_type_embeds

        hidden_states = self.drop(hidden_states)

        output_shape = input_shape + (hidden_states.size(-1),)

        presents = () if use_cache else None
        all_self_attentions = () if output_attentions else None
        all_hidden_states = () if output_hidden_states else None
        for i, (block, layer_past) in enumerate(zip(self.h, past_key_values)):

            # Model parallel
            if self.model_parallel:
                torch.cuda.set_device(hidden_states.device)
                # Ensure layer_past is on same device as hidden_states (might not be correct)
                if layer_past is not None:
                    layer_past = tuple(
                        past_state.to(hidden_states.device) for past_state in layer_past
                    )
                # Ensure that attention_mask is always on the same device as hidden_states
                if attention_mask is not None:
                    attention_mask = attention_mask.to(hidden_states.device)
                if isinstance(head_mask, torch.Tensor):
                    head_mask = head_mask.to(hidden_states.device)
            if output_hidden_states:
                all_hidden_states = all_hidden_states + (hidden_states,)

            if getattr(self.config, "gradient_checkpointing", False) and self.training:

                if use_cache:
                    logger.warning(
                        "`use_cache=True` is incompatible with `config.gradient_checkpointing=True`. Setting "
                        "`use_cache=False`..."
                    )
                    use_cache = False

                def create_custom_forward(module):
                    def custom_forward(*inputs):
                        # None for past_key_value
                        return module(*inputs, use_cache, output_attentions)

                    return custom_forward

                outputs = torch.utils.checkpoint.checkpoint(
                    create_custom_forward(block),
                    hidden_states,
                    None,
                    attention_mask,
                    head_mask[i],
                )
            else:
                outputs = block(
                    hidden_states,
                    layer_past=layer_past,
                    attention_mask=attention_mask,
                    head_mask=head_mask[i],
                    use_cache=use_cache,
                    output_attentions=output_attentions,
                )

            hidden_states = outputs[0]
            if use_cache is True:
                presents = presents + (outputs[1],)

            if output_attentions:
                all_self_attentions = all_self_attentions + (
                    outputs[2 if use_cache else 1],
                )

            # Model Parallel: If it's the last layer for that device, put things on the next device
            if self.model_parallel:
                for k, v in self.device_map.items():
                    if i == v[-1] and "cuda:" + str(k) != self.last_device:
                        hidden_states = hidden_states.to("cuda:" + str(k + 1))

        hidden_states = self.ln_f(hidden_states)

        hidden_states = hidden_states.view(*output_shape)
        # Add last hidden state
        if output_hidden_states:
            all_hidden_states = all_hidden_states + (hidden_states,)

        if not return_dict:
            return tuple(
                v
                for v in [
                    hidden_states,
                    presents,
                    all_hidden_states,
                    all_self_attentions,
                ]
                if v is not None
            )

        return BaseModelOutputWithPast(
            last_hidden_state=hidden_states,
            past_key_values=presents,
            hidden_states=all_hidden_states,
            attentions=all_self_attentions,
        )


class CodeGenForCausalLM(CodeGenPreTrainedModel):
    _keys_to_ignore_on_load_missing = [
        r"h\.\d+\.attn\.masked_bias",
        r"h\.\d+\.attn\.bias",
        r"lm_head\.weight",
    ]

    def __init__(self, config):
        super().__init__(config)
        self.transformer = CodeGenModel(config)
        self.lm_head = nn.Linear(config.n_embd, config.vocab_size)
        self.init_weights()

        # Model parallel
        self.model_parallel = False
        self.device_map = None

    def parallelize(self, device_map=None):
        self.device_map = (
            get_device_map(len(self.transformer.h), range(torch.cuda.device_count()))
            if device_map is None
            else device_map
        )
        assert_device_map(self.device_map, len(self.transformer.h))
        self.transformer.parallelize(self.device_map)
        self.lm_head = self.lm_head.to(self.transformer.first_device)
        self.model_parallel = True

    def deparallelize(self):
        self.transformer.deparallelize()
        self.transformer = self.transformer.to("cpu")
        self.lm_head = self.lm_head.to("cpu")
        self.model_parallel = False
        torch.cuda.empty_cache()

    def get_output_embeddings(self):
        return None

    def set_output_embeddings(self, new_embeddings):
        return

    def prepare_inputs_for_generation(self, input_ids, past=None, **kwargs):
        token_type_ids = kwargs.get("token_type_ids", None)
        # only last token for inputs_ids if past is defined in kwargs
        if past:
            input_ids = input_ids[:, -1].unsqueeze(-1)
            if token_type_ids is not None:
                token_type_ids = token_type_ids[:, -1].unsqueeze(-1)

        attention_mask = kwargs.get("attention_mask", None)
        position_ids = kwargs.get("position_ids", None)

        if attention_mask is not None and position_ids is None:
            # create position_ids on the fly for batch generation
            position_ids = attention_mask.long().cumsum(-1) - 1
            position_ids.masked_fill_(attention_mask == 0, 1)
            if past:
                position_ids = position_ids[:, -1].unsqueeze(-1)
        else:
            position_ids = None
        return {
            "input_ids": input_ids,
            "past_key_values": past,
            "use_cache": kwargs.get("use_cache"),
            "position_ids": position_ids,
            "attention_mask": attention_mask,
            "token_type_ids": token_type_ids,
        }

    def forward(
        self,
        input_ids=None,
        past_key_values=None,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        head_mask=None,
        inputs_embeds=None,
        labels=None,
        use_cache=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
    ):
        r"""
        labels (:obj:`torch.LongTensor` of shape :obj:`(batch_size, sequence_length)`, `optional`):
            Labels for language modeling. Note that the labels **are shifted** inside the model, i.e. you can set
            ``labels = input_ids`` Indices are selected in ``[-100, 0, ..., config.vocab_size]`` All labels set to
            ``-100`` are ignored (masked), the loss is only computed for labels in ``[0, ..., config.vocab_size]``
        """
        return_dict = (
            return_dict if return_dict is not None else self.config.use_return_dict
        )

        transformer_outputs = self.transformer(
            input_ids,
            past_key_values=past_key_values,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )
        hidden_states = transformer_outputs[0]

        # Set device for model parallelism
        if self.model_parallel:
            torch.cuda.set_device(self.transformer.first_device)
            hidden_states = hidden_states.to(self.lm_head.weight.device)

        # make sure sampling in fp16 works correctly and
        # compute loss in fp32 to match with mesh-tf version
        # https://github.com/EleutherAI/gpt-neo/blob/89ce74164da2fb16179106f54e2269b5da8db333/models/gpt2/gpt2.py#L179
        lm_logits = self.lm_head(hidden_states).to(torch.float32)

        loss = None
        if labels is not None:
            # Shift so that tokens < n predict n
            shift_logits = lm_logits[..., :-1, :].contiguous()
            shift_labels = labels[..., 1:].contiguous()
            # Flatten the tokens
            loss_fct = CrossEntropyLoss()
            loss = loss_fct(
                shift_logits.view(-1, shift_logits.size(-1)), shift_labels.view(-1)
            )

            loss = loss.to(hidden_states.dtype)

        if not return_dict:
            output = (lm_logits,) + transformer_outputs[1:]
            return ((loss,) + output) if loss is not None else output

        return CausalLMOutputWithPast(
            loss=loss,
            logits=lm_logits,
            past_key_values=transformer_outputs.past_key_values,
            hidden_states=transformer_outputs.hidden_states,
            attentions=transformer_outputs.attentions,
        )

    @staticmethod
    def _reorder_cache(
        past: Tuple[Tuple[torch.Tensor]], beam_idx: torch.Tensor
    ) -> Tuple[Tuple[torch.Tensor]]:
        """
        This function is used to re-order the :obj:`past_key_values` cache if
        :meth:`~transformers.PretrainedModel.beam_search` or :meth:`~transformers.PretrainedModel.beam_sample` is
        called. This is required to match :obj:`past_key_values` with the correct beam_idx at every generation step.
        """
        return tuple(
            tuple(
                past_state.index_select(0, beam_idx.to(past_state.device))
                for past_state in layer_past
            )
            for layer_past in past
        )



from openelm.environments.sodaracer.walker import Walker
from openelm.environments.sodaracer.walker.CPPN_fixed import (
    make_walker as make_walker_cppn_fixed,
)
from openelm.environments.sodaracer.walker.CPPN_mutable import (
    make_walker as make_walker_cppn_mutable,
)
from openelm.environments.sodaracer.walker.radial import make_walker as make_walker_radial
from openelm.environments.sodaracer.walker.square import make_walker as make_walker_square


def test_seed_walkers():
    """Test that the Sodarace walkers build correctly."""
    # Test that the walkers build correctly
    square_walker: Walker = make_walker_square()
    radial_walker: Walker = make_walker_radial()
    cppn_fixed_walker: Walker = make_walker_cppn_fixed()
    cppn_mutable_walker: Walker = make_walker_cppn_mutable()

    # Test that the walkers have the correct number of joints and muscles
    assert (len(square_walker.joints), len(square_walker.muscles)) == (5, 8)
    assert (len(radial_walker.joints), len(radial_walker.muscles)) == (9, 16)
    assert (len(cppn_fixed_walker.joints), len(cppn_fixed_walker.muscles)) == (24, 96)
    assert (len(cppn_mutable_walker.joints), len(cppn_mutable_walker.muscles)) == (
        24,
        96,
    )

    # Test muscle loading
    correct_distance_joint = [
        0,
        1,
        {"type": "distance", "amplitude": 0.0, "phase": 0.0},
    ]
    correct_muscle = [2, 4, {"type": "muscle", "amplitude": 2.0, "phase": 0.0}]
    assert square_walker.muscles[0] == correct_distance_joint
    assert square_walker.muscles[-1] == correct_muscle

    # Test full dictionary
    correct_square_dict: dict = {
        "joints": [(0, 0), (0, 10), (10, 10), (10, 0), (5, 5)],
        "muscles": [
            [0, 1, {"type": "distance", "amplitude": 0.0, "phase": 0.0}],
            [1, 2, {"type": "distance", "amplitude": 0.0, "phase": 0.0}],
            [2, 3, {"type": "distance", "amplitude": 0.0, "phase": 0.0}],
            [3, 0, {"type": "distance", "amplitude": 0.0, "phase": 0.0}],
            [3, 4, {"type": "distance", "amplitude": 0.0, "phase": 0.0}],
            [0, 4, {"type": "muscle", "amplitude": 5.0, "phase": 0.0}],
            [1, 4, {"type": "muscle", "amplitude": 10.0, "phase": 0.0}],
            [2, 4, {"type": "muscle", "amplitude": 2.0, "phase": 0.0}],
        ],
    }
    assert square_walker.to_dict() == correct_square_dict
    # Test validation
    assert square_walker.validate()
    assert radial_walker.validate()
    assert cppn_fixed_walker.validate()
    assert cppn_mutable_walker.validate()



import random
import re

import torch
from transformers import GPT2TokenizerFast

from openelm.codegen.modelling_codegen import CodeGenForCausalLM
from openelm.constants import PROJECT_PATH


def set_seed(seed, deterministic=True):
    random.seed(seed)
    # os.environ['PYTHONHASHSEED'] = str(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.backends.cudnn.deterministic = deterministic
        torch.backends.cudnn.benchmark = not deterministic
        # torch.use_deterministic_algorithms(deterministic)


def create_model(ckpt_path, fp16=True):
    if fp16:
        return CodeGenForCausalLM.from_pretrained(
            ckpt_path, torch_dtype=torch.float16, low_cpu_mem_usage=True
        )
    else:
        return CodeGenForCausalLM.from_pretrained(ckpt_path)


def create_tokenizer():
    t = GPT2TokenizerFast.from_pretrained("gpt2")
    t.max_model_input_sizes["gpt2"] = 1e20
    return t


def include_whitespace(t, n_min=2, n_max=20, as_special_tokens=False):
    t.add_tokens(
        [" " * n for n in reversed(range(n_min, n_max))],
        special_tokens=as_special_tokens,
    )
    return t


def include_tabs(t, n_min=2, n_max=20, as_special_tokens=False):
    t.add_tokens(
        ["\t" * n for n in reversed(range(n_min, n_max))],
        special_tokens=as_special_tokens,
    )
    return t


def create_custom_gpt2_tokenizer():
    t = create_tokenizer()
    t = include_whitespace(t=t, n_min=2, n_max=32, as_special_tokens=False)
    t = include_tabs(t=t, n_min=2, n_max=10, as_special_tokens=False)
    return t


def truncate(completion, def_num=1, print_num=1, only_local_scope=False):
    def find_re(string, pattern, start_pos):
        m = pattern.search(string, start_pos)
        return m.start() if m else -1

    terminals = [
        re.compile(r, re.MULTILINE)
        for r in ["^#", re.escape("<|endoftext|>"), "^'''", '^"""', "\n\n\n"]
    ]
    prints = list(re.finditer("^print", completion, re.MULTILINE))
    if len(prints) > print_num:
        completion = completion[: prints[print_num].start()]

    if only_local_scope:
        global_lines = list(re.finditer("^[a-zA-Z]", completion, re.MULTILINE))
        if global_lines:
            completion = completion[: global_lines[0].start()]
    else:
        defs = list(re.finditer("^def", completion, re.MULTILINE))
        if len(defs) > def_num:
            completion = completion[: defs[def_num].start()]

    start_pos = 0

    terminals_pos = [
        pos
        for pos in [find_re(completion, terminal, start_pos) for terminal in terminals]
        if pos != -1
    ]
    if len(terminals_pos) > 0:
        return completion[: min(terminals_pos)]
    else:
        return completion


def model_setup(cfg):
    set_seed(cfg.seed, deterministic=True)
    device = torch.device("cuda" if cfg.cuda else "cpu")
    use_fp16 = True
    if not cfg.fp16 or device.type == "cpu":
        use_fp16 = False

    if cfg.model.startswith("codegen-16B"):
        use_fp16 = True

    tokenizer = create_custom_gpt2_tokenizer()
    tokenizer.padding_side = "left"
    tokenizer.pad_token = cfg.pad_token

    ckpt_path = PROJECT_PATH / cfg.checkpoints_dir / cfg.model
    if cfg.gpus > 1:
        model = torch.nn.DataParallel(
            create_model(ckpt_path, fp16=use_fp16), device_ids=list(range(cfg.gpus))
        ).to(device)
    else:
        model = create_model(ckpt_path, fp16=use_fp16).to(device)
    return model, tokenizer


def sample(cfg, model, tokenizer, batch, add_def=False):
    """Run a model on a batch of contexts for a particular task."""
    device = torch.device("cuda" if cfg.cuda else "cpu")

    input_ids_len = batch["input_ids"].shape[1]
    assert input_ids_len < cfg.gen_max_len
    with torch.no_grad():
        batch = batch.to(device)
        if cfg.gpus > 1:
            tokens = model.module.generate(
                **batch,
                do_sample=True,
                num_return_sequences=cfg.batch_size,
                temperature=cfg.temp,
                max_length=input_ids_len + cfg.gen_max_len,
                top_p=cfg.top_p,
                pad_token_id=cfg.pad_token,
                use_cache=True,
            )
        else:
            tokens = model.generate(
                **batch,
                do_sample=True,
                num_return_sequences=cfg.batch_size,
                temperature=cfg.temp,
                max_length=input_ids_len + cfg.gen_max_len,
                top_p=cfg.top_p,
                pad_token_id=cfg.pad_token,
                use_cache=True,
            )
        # "input_ids_len:" removes the prompt
        # - 1 adds in "def"
        if add_def:
            input_ids_len -= 1
        text = tokenizer.batch_decode(tokens[:, input_ids_len:, ...])
    return text



# This file steals codes from https://github.com/moshesipper/tiny_gp and modifies it into one of ELM benchmarks.
# Original header below:

# tiny genetic programming plus, by © moshe sipper, www.moshesipper.com
# graphic output, dynamic progress display, bloat-control option
# need to install https://pypi.org/project/graphviz/

import itertools
import sys
from copy import deepcopy
from random import randint, random
from statistics import mean
from typing import Callable, Iterable, Optional, Union

from graphviz import Digraph
from IPython.display import Image, display

MIN_DEPTH = 0  # minimal initial random tree depth
PROB_MUTATION = 0.2  # per-node mutation probability
NUM_MUTATION_PER_LOOP = (
    10000  # the number of mutations to be run per step of the experiments
)


def add(x, y):
    return x + y


def mod(x, y):
    return x % y


def mul(x, y):
    return x * y


def subtract(x, y):
    return x - y


def four_parity_reference(b1, b2, b3, b4):
    bit_sum = sum([b1, b2, b3, b4])
    return bit_sum % 2


def quadratic_references(a, b, c, x):
    return a * pow(x, 2) + b * x + c


def generate_four_parity():
    """
    Hand-coded four_parity.
    Returns:
        the GPTree object that implements four_parity.
    """
    FUNCTIONS = [add, mod]
    TERMINALS = (
        [f"b{i}" for i in range(1, 5)]
        + [f"c{i}" for i in range(1, 5)]
        + [-2, -1, 0, 1, 2]
    )
    args = (FUNCTIONS, TERMINALS)
    root = GPTree(*args, data=mod, right=GPTree(*args, data=2))
    prev_node = GPTree(
        *args, data=add, left=GPTree(*args, data="b3"), right=GPTree(*args, data="b4")
    )
    for i in range(2, 0, -1):
        node = GPTree(
            *args, data=add, left=GPTree(*args, data=f"b{i}"), right=prev_node
        )
        prev_node = node
    root.left = prev_node

    return root


def generate_quadratic():
    """
    Hand-coded quadratic.
    Returns:
        the GPTree object that implements quadratic.
    """
    FUNCTIONS = [add, subtract, mul, pow]
    TERMINALS = ["a", "b", "c", "x"] + [-2, -1, 0, 1, 2]
    args = (FUNCTIONS, TERMINALS)
    root = GPTree(*args, data=add, right=GPTree(*args, data="c"))

    node_1 = GPTree(
        *args, data=pow, left=GPTree(*args, data="x"), right=GPTree(*args, data=2)
    )
    node_2 = GPTree(*args, data=mul, left=GPTree(*args, data="a"), right=node_1)
    node_3 = GPTree(
        *args,
        data=add,
        left=node_2,
        right=GPTree(
            *args, data=mul, left=GPTree(*args, data="b"), right=GPTree(*args, data="x")
        ),
    )
    root.left = node_3

    return root


def generate_dataset(inputs, target_fn=four_parity_reference):
    ground_truth = [target_fn(*i) for i in inputs]
    return inputs, ground_truth


class GPTree:
    def __init__(self, functions, terminals, data=None, left=None, right=None):
        self.functions = functions
        self.terminals = terminals
        self.data = data
        self.left = left
        self.right = right

    def node_label(self):  # return string label
        if self.data in self.functions:
            return self.data.__name__
        else:
            return str(self.data)

    def draw(self, dot, count):  # dot & count are lists in order to pass "by reference"
        node_name = str(count[0])
        dot[0].node(node_name, self.node_label())
        if self.left:
            count[0] += 1
            dot[0].edge(node_name, str(count[0]))
            self.left.draw(dot, count)
        if self.right:
            count[0] += 1
            dot[0].edge(node_name, str(count[0]))
            self.right.draw(dot, count)

    def draw_tree(self, fname, footer):
        dot = [Digraph(fname, comment=footer)]
        count = [0]
        self.draw(dot, count)
        dot[0].format = "png"
        dot[0].render(directory="")
        display(Image(filename=f"{fname}.gv.png"))
        return dot[0].source

    def compute_tree(self, b, arg_names=("b1", "b2", "b3", "b4")):
        """
        Args:
            b: a list/tuple of inputs (b1, b2, b3, b4)
            arg_names: argument names.
        Returns:
            the evaluation at this node.
        """
        if not isinstance(b, (list, tuple)):
            raise TypeError(f"Input b must be a list or tuple. Got {type(b)} instead.")

        arg_dict = {name: value for name, value in zip(arg_names, b)}

        if self.data in self.functions:
            return self.data(self.left.compute_tree(b), self.right.compute_tree(b))
        elif isinstance(self.data, str):
            return arg_dict[
                self.data
            ]  # it will take care of errors like unrecognized variable 'c1', etc.
        else:
            return self.data  # self.data is a number

    def random_tree(
        self, grow, max_depth, depth=0
    ):  # create random tree using either grow or full method
        if depth < MIN_DEPTH or (depth < max_depth and not grow):
            self.data = self.functions[randint(0, len(self.functions) - 1)]
        elif depth >= max_depth:
            self.data = self.terminals[randint(0, len(self.terminals) - 1)]
        else:  # intermediate depth, grow
            if random() > 0.5:
                self.data = self.terminals[randint(0, len(self.terminals) - 1)]
            else:
                self.data = self.functions[randint(0, len(self.functions) - 1)]
        if self.data in self.functions:
            self.left = GPTree(self.functions, self.terminals)
            self.left.random_tree(grow, max_depth, depth=depth + 1)
            self.right = GPTree(self.functions, self.terminals)
            self.right.random_tree(grow, max_depth, depth=depth + 1)

    def mutation(self):
        if random() < PROB_MUTATION:  # mutate at this node
            self.random_tree(grow=True, max_depth=2)
        elif self.left:
            self.left.mutation()
        elif self.right:
            self.right.mutation()

    def size(self):  # tree size in nodes
        if self.data in self.terminals:
            return 1
        left_size = self.left.size() if self.left else 0
        right_size = self.right.size() if self.right else 0
        return 1 + left_size + right_size

    def build_subtree(self):  # count is list in order to pass "by reference"
        t = GPTree(self.functions, self.terminals)
        t.data = self.data
        if self.left:
            t.left = self.left.build_subtree()
        if self.right:
            t.right = self.right.build_subtree()
        return t

    def scan_tree(
        self, count, second
    ):  # note: count is list, so it's passed "by reference"
        count[0] -= 1
        if count[0] <= 1:
            if not second:  # return subtree rooted here
                return self.build_subtree()
            else:  # glue subtree here
                self.data = second.data
                self.left = second.left
                self.right = second.right
        else:
            ret = None
            if self.left and count[0] > 1:
                ret = self.left.scan_tree(count, second)
            if self.right and count[0] > 1:
                ret = self.right.scan_tree(count, second)
            return ret


def error(individual, dataset):
    return mean([abs(individual.compute_tree(ds[0]) - ds[1]) for ds in dataset])


def swap_node(
    tree: Optional[GPTree],
    tree_data: Union[str, Callable],
    target_data: Union[str, Callable],
) -> bool:
    """
    Swap the name of a variable into another one (only apply to the first encounter of a DFS).
    Args:
        tree: the GPTree node.
        tree_data: the variable name or the function.
        target_data: the target variable name or the function.
    Returns:
        True if the variable is found and replaced.
    """
    if tree is None:
        return False

    if tree.data == tree_data:
        tree.data = target_data
        return True

    if not swap_node(tree.left, tree_data, target_data):
        return swap_node(tree.right, tree_data, target_data)
    return True


def eval_tree(tree: GPTree, dataset: Iterable) -> list:
    """
    Test the correctness of a GPTree against a dataset.
    Args:
        tree: the tree to test against.
        dataset: (inputs, ground_truth)
    Returns
        the list of evaluation results.
        (I'm trying to match the error codes here:
            0: pass.
            1: error in code, but runs.
            2: fails to run, e.g., when it has unrecognized variable name like 'c1'.)
    """
    results = []
    for data in zip(*dataset):
        try:
            output = tree.compute_tree(data[0])
            results.append(
                0 if output == data[1] else 1
            )  # right or wrong, but no error.
        except Exception:
            results.append(2)  # Fails to run.

    return results


def list_equal(l1, l2):
    return all([x == y for x, y in zip(l1, l2)])


def mutate_compare(
    tree: GPTree, num_mutation: int, dataset: tuple
) -> tuple[float, float]:
    """
    Mutate (a copy) of the tree num_mutation times, and return the percentage of successful mutations.
    Args:
        tree: the tree to mutate (will make a copy before mutation).
        num_mutation: number of times to mutate.
        dataset: the dataset to test against. Format: (input, ground_truth) where input and ground_truth are lists
    Returns:
        success rate, wrong answer rate (without error)
    """
    corrected, wrong_answers = 0, 0
    for j in range(num_mutation):
        tree_copy = deepcopy(tree)
        tree_copy.mutation()
        eval_result = eval_tree(tree_copy, dataset)
        if list_equal(eval_result, [0] * len(dataset[1])):
            corrected += 1
        else:
            wrong_answers += not (2 in eval_result)

    return corrected / num_mutation, wrong_answers / num_mutation


def draw_tree(tree: GPTree, name: str, options: list):
    if any([opt in sys.argv[1:] for opt in options]):
        tree.draw_tree(name, "")
        print(
            f"(if running on terminal, check out {name}.gv.png for the tree structure)"
        )


def main():
    tree = generate_four_parity()
    four_parity_dataset = generate_dataset(
        [i for i in itertools.product(range(2), repeat=4)]
    )

    # According to the paper, the error is gradually introduced by replacing b-variables to c-variables step-by-step.
    # At the very end, modulo 2 is replaced by modulo 3. We carry the experiments out here and collect the percentage
    #   of mutations that fix the problem.

    print(
        "Start of experiments "
        "(use -v to draw the original trees, or use --all to draw all the mutated trees)."
    )

    print("Four parity:")
    draw_tree(tree, "four_parity", ["-v", "--all"])

    success_rate, wrong_answer_rate = mutate_compare(
        tree, NUM_MUTATION_PER_LOOP, four_parity_dataset
    )
    print(
        f"0 bug, successful mutation rate: {success_rate}; "
        f"wrong answer rate (executed without error): {wrong_answer_rate}."
    )
    for i in range(4):
        swap_node(tree, f"b{i+1}", f"c{i+1}")
        draw_tree(tree, f"four_parity_{i+1}_bug", ["--all"])
        success_rate, wrong_answer_rate = mutate_compare(
            tree, NUM_MUTATION_PER_LOOP, four_parity_dataset
        )
        print(
            f"{i+1} bug, successful mutation rate: {success_rate}; "
            f"wrong answer rate (executed without error): {wrong_answer_rate}"
        )

    # The paper doesn't seem to specify what they did with `quadratic`. So here is my own version.

    tree = generate_quadratic()
    quadratic_dataset = generate_dataset(
        [i for i in itertools.product(range(2), repeat=4)],
        target_fn=quadratic_references,
    )

    print("Quadratic:")
    draw_tree(tree, "quadratic", ["-v", "--all"])

    success_rate, wrong_answer_rate = mutate_compare(
        tree, NUM_MUTATION_PER_LOOP, quadratic_dataset
    )
    print(
        f"0 bug, successful mutation rate: {success_rate}; "
        f"wrong answer rate (executed without error): {wrong_answer_rate}."
    )
    for i in range(2):
        swap_node(tree, add, subtract)
        draw_tree(tree, f"quadratic_{i+1}_bug", ["--all"])
        success_rate, wrong_answer_rate = mutate_compare(
            tree, NUM_MUTATION_PER_LOOP, quadratic_dataset
        )
        print(
            f"{i+1} bug, successful mutation rate: {success_rate}; "
            f"wrong answer rate (executed without error): {wrong_answer_rate}"
        )


if __name__ == "__main__":
    main()



import re
from enum import Enum
from typing import Optional

line_number_pattern = re.compile(r"(?m)^@@ -(?P<l1>\d*),*?(?P<s1>\d*?) \+(?P<l2>\d*),*?(?P<s2>\d*?) @@")
diff_pattern = re.compile(
    r"""<NME> (?P<name>.*?)
<BEF> (?P<file>(.|\n)*?)
<MSG> (?P<message>(.|\n)*?)
<DFF> (?P<diff>(.|\n)*)"""
)
hunk_split_pattern = re.compile(r"(?m)^(@@ .*? @@).*\n")
ignored = re.compile(r"(?m)^\\ No newline at end of file$\n?")


class DiffState(Enum):
    """
    An Enum keeping track of the validity of the diff data. It is the return of the helper function `verify_diff`.
    Binary codes help internally, as some errors are additive (e.g., can have both invalid text and invalid line num).
    But we convert the binary codes into Enum for better readability.
    """

    VALID = 0b000  # valid diff

    # The following are errors that can still be either ignored or fixed.
    INVALID_TEXT = 0b001  # pre-diff texts cannot be found in the context.
    INVALID_LINE_NUM = (
        0b010  # the numbers in @@ -x,y +a,b @@ are invalid (but can be parsed).
    )
    INVALID_TEXT_AND_LINE_NUM = 0b011  # both 0b001 and 0b010.

    # The following are format errors that cannot be ignored.
    BAD_FORMAT = 0b100  # cannot be parsed according to <NME> ...\n<BEF> ...\n<MSG> ...\n<DFF> @@ ... @@\n...
    BAD_DIFF_HUNK_FORMAT = 0b101  # diff hunk contains lines whose initial character is not one of ' ', '+', '-'
    BAD_LINE_NUM_FORMAT = (
        0b110  # the @@ ... @@ bracket can be found but numbers cannot be parsed.
    )
    BAD_HUNK_AND_LINE_FORMAT = 0b111  # both 0b110 and 0b101.


def split_diff(content: str) -> dict:
    """
    Args:
        content: the diff content.

    Returns:
        A dict with potentially 4 items:
            name: the filename
            file: the file content
            message: the diff message
            diff: the diff hunk
        Any key could be missing. That would mean a failure in matching.
    """
    match = diff_pattern.match(content)
    return {} if match is None else match.groupdict()


def parse_line_info(content: str) -> tuple:
    """
    Parse @@ -x,y +a,b @@

    Args:
        the @@ ... @@ line
    Returns:
        (x, y, a, b) as integers
    """
    match = line_number_pattern.match(content)
    if match is None:
        return ()
    match_dict = match.groupdict()
    # line numbers are mandatory
    if not match_dict['l1'] or not match_dict['l2']:
        return ()
    for s in ['s1', 's2']:
        # line ranges are optional and default to 1
        match_dict[s] = match_dict[s] if match_dict[s] else '1'
    return int(match_dict['l1']), int(match_dict['s1']), int(match_dict['l2']), int(match_dict['s2'])


def parse_diff_content(
    hunk: str, separate_lines=False, reject_invalid=False
) -> Optional[tuple]:
    """
    Parse a diff content to turn it into (before_diff, after_diff) based on '+', '-' at the beginning of each line.

    Args:
        hunk: the diff content (without "@@ ... @@").
        separate_lines: (Optional) True if return list of lines.
        reject_invalid: (Optional) True if return None for invalid diff hunk (non-empty lines without starting
            with ' ', '-', '+')
    Returns:
        (before_diff, after_diff);
        None if reject_invalid==True and the diff hunk contains invalid format.
    """
    hunk = hunk.split("\n")
    before_diff, after_diff = [], []
    for line in hunk:
        # Ignore invalid trailing '\n'. An empty line in the diff hunk should at least be '\n ' with the space.
        if not line:
            continue
        if line[0] == "-" or line[0] == " ":
            before_diff.append(line[1:])
        if line[0] == "+" or line[0] == " ":
            after_diff.append(line[1:])
        if reject_invalid:
            if all([line[0] != c for c in [" ", "-", "+"]]):
                return None
    if separate_lines:
        return before_diff, after_diff
    else:
        return "\n".join(before_diff), "\n".join(after_diff)


def replace_text(text: str,
                 before: str,
                 after: str,
                 start_pointer: int,
                 reject_incomplete_line: bool = True) -> tuple[str, int]:
    """
    Try to match `before` within `text` and replace the content into `after`.
    If not found, return the original text.

    Args:
        text: the original text.
        before: the text to be matched.
        after: the text to be replaced into.
        start_pointer: the index where we start to match (inclusive).
        reject_incomplete_line: (Optional) reject the patch if `before` does not match till the end of a line.
    Returns:
        (diff_result, new_start_pointer)
        the text after the match-and-replace and the new index at the end of the change.
    """
    idx = text[start_pointer:].find(before)
    start_idx = start_pointer + idx

    if reject_incomplete_line:
        # If the end of the match is neither EOF nor \n, reject the patch.
        if idx >= 0 and start_idx + len(before) < len(text) and text[start_idx + len(before)] != '\n':
            return text, start_pointer

    if idx < 0:
        return text, start_pointer
    else:
        # Even if start_idx + len(before) is out-of-bound, the list slicing would return ""
        return text[:start_idx] + after + text[start_idx + len(before):], start_idx + len(after)


def apply_diff(file: str, diff: str, use_line_number=False, allow_add_file=True) -> str:
    """
    Apply the diff to the file content. We try to be lenient and keep applying the patch naively until we cannot.
    (Note: use_line_number=True is somehow slightly slower.)
    (Warning: if use_line_number==False, we could have some problematic cases like, if all lines in diff hunk
        starts with "+", the pre-diff paragraphs relevant to the hunk is empty. Because we only use pre-diff
        paragraphs to match, we would simply match the very beginning.)
    Args:
        file: the file content.
        diff: the diff hunk (containing "@@ -x,y +a,b @@").
        use_line_number: (Optional) use the line numbers in "@@ ... @@" faithfully.
        allow_add_file: (Optional) when file is "ADDFILE" (meaning <BEF> ADDFILE\n... showed up in the diff text),
            we automatically patch the diff by a direct replacement.
    Return:
        the maximally patched file content.
    """
    diff = hunk_split_pattern.split(ignored.sub("", diff))
    # If we use the line numbers, we match-and-replace in a line-by-line fashion.
    file_by_line = file.split("\n") if use_line_number else None
    line_offset = 0  # the offset between pre-/post-patching line numbers

    # If we do not use the line numbers, for multiple diff hunk, we only move forward in a greedy manner.
    patch_pointer = 0

    i = (
        0 if diff[0] else 1
    )  # We have delimiter at the beginning, causing empty initial string
    while (
        i < len(diff) - 1
    ):  # Need at least a pair of '@@ ... @@' and diff hunk to continue
        # Expect a string with '@@ ... @@' followed by a diff hunk
        line_info = parse_line_info(diff[i])
        diff_content = diff[i + 1]
        i += 2

        # Generate the pre-/post-diff string based on the first character being '+' or '-'
        # (Note: parse_diff_content will ignore trailing \n at the beginning and at the end)
        parsed_diff = parse_diff_content(diff_content, separate_lines=use_line_number)

        # If we allow the recognition of "ADDFILE", special treatment is needed.
        if allow_add_file and file == "ADDFILE":
            if use_line_number:
                # Immediately apply the first hunk but also check the partial validity of line numbers.
                return parsed_diff[1] if line_info == (0, 0) else ""
            else:
                # Immediately apply the first hunk and ignore the rest.
                return parsed_diff[1]

        if use_line_number:
            # If line numbers cannot be parsed, skip.
            if not line_info:
                continue

            # Offset the starting line
            start_idx = line_info[0] + line_offset

            # Match the referred lines with the file context
            referred_lines = file_by_line[start_idx - 1 : start_idx - 1 + line_info[1]]
            valid = all([l1 == l2 for l1, l2 in zip(parsed_diff[0], referred_lines)])

            # If lines fully match and the number of lines is consistent, apply the patch.
            # We ignore the second pair "+a, b" just to be lenient.
            if valid and len(parsed_diff[0]) == line_info[1]:
                # Update the list of lines
                if start_idx == 0:  # Add lines to the beginning.
                    file_by_line = parsed_diff[1] + file_by_line
                else:
                    file_by_line = file_by_line[: start_idx - 1] + parsed_diff[1] + \
                                   file_by_line[start_idx - 1 + line_info[1]:]
                line_offset += len(parsed_diff[1]) - line_info[1]
        else:
            # CAUTION: this way of handling empty context is being very lenient and could lead to
            # undesirable behaviors. Only do this when you want to be as tolerant as possible.
            if parsed_diff[0] == "":
                if patch_pointer != 0:  # Lack of matching context can only happen at the beginning of file.
                    continue
                file = parsed_diff[1] + "\n" + file
                patch_pointer = len(parsed_diff[0]) + 1
            else:
                # Directly (and naively) apply patch by match-and-replace.
                file, patch_pointer = replace_text(file, parsed_diff[0], parsed_diff[1], patch_pointer)

    if use_line_number:
        file = "\n".join(file_by_line)
    return file


def verify_diff(diff_text: str) -> DiffState:
    """
    Verify the validity of a complete diff text.

    Args:
        diff_text: the complete diff text.
            The overall format conforms "<NME> ...\n<BEF> ...\n<MSG> ...\n<DFF> ..." and the text
            after <DFF> has 1 or more lines of "@@ -x,y +a,b @@" followed by the corresponding hunk.
    Returns:
        A DiffState (see above).
    """
    diff_dict = split_diff(ignored.sub("", diff_text))  # Ignore the GitHub warning on the end of file
    line_offset = 0

    keys = ["name", "file", "message", "diff"]
    for key in keys:
        if key not in diff_dict:
            return DiffState(0b100)  # Invalid overall format

    diff_parts = hunk_split_pattern.split(diff_dict["diff"])
    if not diff_parts:
        return DiffState(0b100)  # Invalid overall format

    context_mismatch, line_number_mismatch = False, False
    bad_diff_hunk, bad_line_number = False, False

    i = 0 if diff_parts[0] else 1
    while (
        i < len(diff_parts) - 1
    ):  # Need at least a pair of '@@ ... @@' and diff hunk to continue
        line_info = parse_line_info(diff_parts[i])
        diff_content = parse_diff_content(diff_parts[i + 1], reject_invalid=True)
        i += 2

        # Special treatment if we are adding a new file
        if diff_dict["file"] == "ADDFILE":
            if (
                len(diff_parts) != i
                or not line_info
                or line_info[:3] != (0, 0, 1)
                or line_info[3] != len(diff_content[1].split("\n"))
                or diff_content[0]
            ):
                return DiffState(0b110)
            else:
                return DiffState(0b000)

        if not line_info or len(line_info) != 4:
            bad_line_number = True
        if diff_content is None:
            bad_diff_hunk = True

        # Skip the diff matching checks if bad format already occurred
        if bad_diff_hunk or bad_line_number:
            continue

        # Try to see if there is a match in the file context. Must match complete lines or till EOF.
        match_idx = diff_dict["file"].find(diff_content[0])
        if match_idx == -1 or (
            match_idx + len(diff_content[0]) != len(diff_dict["file"])
            and diff_dict["file"][match_idx + len(diff_content[0])] != "\n"
        ):
            context_mismatch = True

        if line_info[0] <= 0:
            # -0,0 only happens when we create a new file (in which case the context is <BEF> ADDFILE\n...).
            if line_info[1] != 0 or diff_dict["file"] != "ADDFILE":
                line_number_mismatch = True
        else:
            # Check the line numbers regardless of whether the context matches.
            pre_diff_line_number = len(diff_content[0].split("\n"))
            post_diff_line_number = len(diff_content[1].split("\n"))
            if (pre_diff_line_number, post_diff_line_number) != (
                line_info[1],
                line_info[3],
            ):
                line_number_mismatch = True
            else:
                line_offset += len(diff_content[1]) - line_info[1]

    if bad_diff_hunk or bad_line_number:
        return DiffState(bad_diff_hunk * 0b001 + bad_line_number * 0b010 + 0b100)
    else:
        return DiffState(context_mismatch * 0b001 + line_number_mismatch * 0b010)
